// vincia23-root.cc is a part of the VINCIA event generator.
// Copyright (C) 2016 Peter Skands.
// VINCIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This is a simple test program (adapted from the PYTHIA example main05.cc)
// It studies Z bosons at the LHC, summing over all contributing channels

// PYTHIA and VINCIA include files and namespaces
#include "Vincia/Vincia.h"
using namespace Pythia8;
using namespace Vincia;

// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
#include "VinciaRoot/PythiaRoot.h"

// Global on/off switches for optional additional plots:
// 1) pTin and pTout distributions
const bool OPTcreateDisplays = true;

int main() {

  // Generator. LHC process and output selection. Initialization.
  Pythia pythia;
  VinciaPlugin vincia(&pythia,"vincia23.cmnd");

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  
  // Extract settings to be used in the main program.  
  double eCM              = settingsPtr->parm("Beams:eCM");
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");  

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");

  // MPI and hadronization settings
  bool   mpi              = settingsPtr->flag("PartonLevel:MPI");
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";

  //
  // INITIALIZE
  //
  vincia.init();

  // Define a few simple PYTHIA Histograms.
  Hist HmZ("mZ", 100,50.,150.);
  Hist HmHat("mHat",100,0.,1000.);
  
  //************************************************************************
  // VinciaRoot
  // Create VinciaRoot object and set plot labels, legends, etc
  // Also define names of files to save and read histograms to/from
  //************************************************************************

  // Create VinciaRoot object and set default behavior to logarithmic y axes
  VinciaRoot vroot; 
  vroot.setLogY();
  // Set range on ratio plots (default is [0.5,1.5])
  vroot.setRatioRange("all",0.5,1.5);
  vroot.setShowLegendChi2("all",false);
  // Do not show uncertainty breakdowns
  vroot.setShowRelUnc("all",false);
  
  // Optionally set left/right white-on-black labels
  vroot.setLeftLabel("all","pp#rightarrowZ#rightarrowleptons");
  string eCMstring = num2str(int(eCM)) + " GeV";
  vroot.setRightLabel("all",eCMstring);

  // Set first update and save event
  double iSave   = 100;
  
  // Set MC names and versions
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  string mcVersions = "Vincia " + vcVer + " + Pythia "+pyVer;
  
  // Save current run to file with default name:
  string mcFileSave = "vincia23";

  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia23").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia23").c_str());
  
  // Examples of how to set plot information and define read/save file names
  string mcLabel;
  // Example 1: PYTHIA runs
  if (!vinciaOn) {
    mcLabel    = "Pythia";
  }
  // Example 2: VINCIA runs
  else {
    mcLabel    = "Vincia";
  }
  // Example 3: check if special settings results (hadronization or MPI OFF)
  if (!hadronLevel) {
    mcLabel += " partons";
    mcFileSave += "-noHad";
  }
  if (!mpi) {
    mcLabel += " no MPI";
    mcFileSave += "-noMpi";
  }

  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend("all",mcLabel);
  vroot.setMCnames("all",mcVersions);
  
  // Book VinciaROOT histograms

  // pTZ tail spectrum
  int    nBins  = 396;
  double xLo    = 8.0;
  double xHi    = 800.0;
  vector<TH1D*> VpTZtail;
  VpTZtail.push_back(new TH1D("pTZtail","p_{TZ} (Born)",nBins,xLo,xHi));
  double wpTZtail = nBins/(xHi-xLo);
  VpTZtail[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZtail[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZtail[0]->SetMinimum(0.0000001);
  VpTZtail[0]->SetMaximum(10.);
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZtail",VpTZtail,400,500);
    vroot.setDataFile("pTZtail",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZtail",mcFileRead[iRead]+"-pTZtail.dat");
    }
    vroot.setRebinLikeData("pTZtail");
    vroot.setLogX("pTZtail");
  } 
 
  // Spectrum after the first recoil
  vector<TH1D*> VpTZtailOne;
  VpTZtailOne.push_back(new TH1D("pTZtailOne","p_{TZ} (Born)",nBins,xLo,xHi));
  double wpTZtailOne = nBins/(xHi-xLo);
  VpTZtailOne[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZtailOne[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZtailOne[0]->SetMinimum(0.0000001);
  VpTZtailOne[0]->SetMaximum(10.);
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZtailOne",VpTZtailOne,400,500);
    vroot.setDataFile("pTZtailOne",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("pTZtailOne",mcFileRead[iRead]+"-pTZtailOne.dat");
    vroot.setRebinLikeData("pTZtailOne");
    vroot.setLogX("pTZtailOne");
  }
  
  // pTZ Peak spectrum
  nBins = 63.0;
  xLo =   0.0;
  xHi =  42.0;
  vector<TH1D*> VpTZpeak;
  VpTZpeak.push_back(new TH1D("pTZpeak","p_{TZ} (Born)",nBins,xLo,xHi));
  double wpTZpeak = nBins/(xHi-xLo);
  VpTZpeak[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZpeak[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZpeak[0]->SetMinimum(-0.02);
  VpTZpeak[0]->SetMaximum(0.095);
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZpeak",VpTZpeak,400,500);
    vroot.setDataFile("pTZpeak",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZpeak",mcFileRead[iRead]+"-pTZpeak.dat");
    }
    vroot.setRebinLikeData("pTZpeak");
  }
  
  // spectrum after the first recoil
  vector<TH1D*> VpTZpeakOne;
  VpTZpeakOne.push_back(new TH1D("pTZpeakOne","p_{TZ} (Born)",nBins,xLo,xHi));
  double wpTZpeakOne = nBins/(xHi-xLo);
  VpTZpeakOne[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZpeakOne[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZpeakOne[0]->SetMinimum(-0.02);
  VpTZpeakOne[0]->SetMaximum(0.095);
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZpeakOne",VpTZpeakOne,400,500);
    vroot.setDataFile("pTZpeakOne",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZpeakOne",mcFileRead[iRead]+"-pTZpeakOne.dat");
    }
  }
  
  // Weights
  xLo = -1.1;
  xHi =  7.1;
  double binWid = 0.02;
  nBins = (xHi-xLo)/binWid;
  vector<TH1D*> Vweights;
  Vweights.push_back(new TH1D("weights","Weights",nBins,xLo,xHi));
  double wtWeights = 1./binWid;
  Vweights[0]->GetXaxis()->SetTitle("weight");
  Vweights[0]->GetYaxis()->SetTitle("N");
  Vweights[0]->SetMinimum(1e-4);
  Vweights[0]->SetMaximum(100.0);
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("weights",Vweights,400,300);
    vroot.setNormalize("weights");
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("weights",mcFileRead[iRead]+"-weights.dat");
    }
  }

  //************************************************************************
  // Event Loop
  //************************************************************************
  
  // Begin event loop. Generate event. Skip if error.
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {
    if (!pythia.next()) continue;

    // Reference to event
    Event& event=pythia.event;

    double wt = vincia.weight();
    HmHat.fill(pythia.info.mHat(),wt);
    HmZ.fill(pythia.process[5].m(),wt);
    Vweights[0]->Fill(wt,wtWeights);
    
    // Find the first and last Z boson
    int iZ = 1;
    int iZ1 = 0;
    for (int i=5; i<pythia.event.size(); ++i)
      if (pythia.event[i].id() == 23) {
	iZ = i; 
	if (iZ1 == 0 && pythia.event[i].pT() > 0.001
	    && abs(pythia.event[i].status()) != 62) iZ1 = i;
      }
    
    // Check daughter lepton momenta
    int ida1 = event[iZ].daughter1();
    int ida2 = event[iZ].daughter2();
    if (abs(event[ida1].eta()) > 2.4 || event[ida1].pT() < 20.) continue;
    if (abs(event[ida2].eta()) > 2.4 || event[ida2].pT() < 20.) continue;    
    // Plot pTZ
    double pTZborn = pythia.event[iZ].pT();
    VpTZpeak[0]->Fill(pTZborn,wpTZpeak*wt);
    VpTZtail[0]->Fill(pTZborn,wpTZtail*wt);

    // Plot pTZ after first emission
    double pTZ1 = pythia.event[iZ1].pT();
    if (iZ1 == 0) pTZ1 = 0.001;
    VpTZpeakOne[0]->Fill(pTZ1,wpTZpeakOne*wt);
    VpTZtailOne[0]->Fill(pTZ1,wpTZtailOne*wt);

    // Periodically save displays
    if (iEvent + 0.5 >= int(iSave)) {
      iSave *= sqrt(2);
      iSave = min(iSave,double(nEvent));
      // Save displays with base name mcFile and file type .pdf 
      // (it is possible to choose other formats, like eps or png)
      vroot.updateDisplays();       
      vroot.saveDisplays(mcFileSave,"pdf");
      // Also save histograms as dat files
      vroot.saveHistograms(mcFileSave,"dat");
      cout<<"Next save at event "<<int(iSave + 0.5)<<endl;
    }
  }  

  // End-of-run update and save
  vroot.updateDisplay();       
  vroot.saveDisplays(mcFileSave, "pdf");
  vroot.saveHistograms(mcFileSave,"dat");

  pythia.stat();
  cout<<HmHat<<HmZ;

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;

  // Done.
  return 0;
}
