// SpaceAntShower.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the SpaceAntennaShower class
  
#ifndef VINCIASPACEANTSHOWER_H
#define VINCIASPACEANTSHOWER_H

#include "Pythia8/SpaceShower.h"
#include "Vincia/VinciaCommon.h"
#include "Vincia/VinciaWeights.h"

#include "Vincia/AntennaSets.h"
#include "Vincia/Branchelementals.h"
#include "Vincia/Reconnections.h"
#include "Vincia/Resolution.h"
#include "Vincia/TrialGeneratorsISR.h"
#include "Vincia/MEcorrections.h"
#include "Vincia/VinciaHooks.h"

namespace Vincia {
  
  // Forward declaration of time shower
  class TimeAntShower;
  
  // Classes contained in this file 
  class SpaceAntShower; 
  
  //**************************************************************************

  // The SpaceAntShower class: main shower class for the timelike antenna shower   
  // Inherits from SpaceShower in Pythia 8 so can be used as alternative to 
  // SpaceShower. Added by NF.
  
  class SpaceAntShower : public Pythia8::SpaceShower {
    
    // Allow VinciaPlugin to access private information
    friend class VinciaPlugin;
    
  public:
   
    // ----------------------------------------------------------------------
    
    // CONSTRUCTOR AND DESTRUCTOR 
    
    // Constructor.
    SpaceAntShower() : isInit(false), headerIsPrinted(false) { };
  
    // Destructor.
    virtual ~SpaceAntShower() { };
  
    // -----------------------------------------------------------------------
    
    // FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING 

    // Initialize generation. Possibility to force re-initialization by hand.
    virtual void init( BeamParticle* beamAPtrIn, BeamParticle* beamBPtrIn);

    // Find whether to limit maximum scale of emissions.
    virtual bool limitPTmax( Event& /*event*/, double /*Q2Fac = 0.*/,
                             double /*Q2Ren = 0.*/) {
      // We handle scales internally, unless (external) merging
      return doMergeSav;
    }

    // Potential enhancement factor of pTmax scale for hardest emission.
    // Used if limitPTmax = true.
    virtual double enhancePTmax() const {
      return 1.0;
    }

    // Prepare system for evolution; identify ME.
    virtual void prepare( int iSys, Event& event, bool limitPTmaxIn = false);

    // Update dipole list after each ISR emission.
    virtual void update( int iSys, Event& event, bool hasWeakRad = false);

    // Select next pT in downwards evolution.
    virtual double pTnext( Event& event, double pTbegAll, double pTendAll,
                           int nRadIn = -1, bool doTrialIn = false);

    // ME corrections and kinematics that may give failure.
    virtual bool branch( Event& event);
    
    // Utility to print dipole list; for debug mainly.
    virtual void list( ostream& os = cout ) const;
    
    // Functions inherited for merging

    // Return clustering kinematics - as needed for merging.
    virtual Event clustered( const Event&, int, int, int, string);
    
    // Return the evolution variable.
    virtual vector<double> getStateVariables ( const Event&, int, int, int, string);
    
    // Check if attempted clustering is handled by spacelike shower
    virtual bool isSpacelike( const Event&, int, int, int, string);
    
    // Return a string identifier of a splitting.
    virtual string getSplittingName( const Event&, int, int, int);
    
    // Return the splitting probability.
    virtual double getSplittingProb( const Event&, int, int, int, string);
    
    // END OF FUNCTIONS INHERITED FROM PYTHIA 8 FOR INTERFACING

    // -----------------------------------------------------------------------
    
    // PUBLIC NON-INHERITED MEMBERS
    
    // Update antennae if FSR changed system in prepare
    void updateAfterFSR(int iSys, Event& event, map<int,int> iChanged);
    
    // Wrapper function to return a specific antenna inside a specific antenna set
    // 0: User, 1: Def, 2: Max, 3: Min    
    AntennaISR* getAnt(int iAnt, int iSet=0) {
      return antennaSets[iSet]->getAnt(iAnt);
    }
    
    // Initialize Pointers to Antenna sets
    void initAntPtr(AntennaSetISR* antUSR, AntennaSetISR* antDEF=0, 
                    AntennaSetISR* antMAX=0, AntennaSetISR* antMIN=0) {
      antennaSets.resize(0);
      antennaSets.push_back(antUSR);
      if (antDEF != 0 && antMAX != 0 && antMIN != 0) {
        antennaSets.push_back(antDEF);
        antennaSets.push_back(antMAX);
        antennaSets.push_back(antMIN);
      }
    }
    
    // Print out the antennae
    void printAnts(int iSys);
    
    // Print header information (version, settings, parameters, etc)
    void header();
    
    // Print final statistics information
    void printInfo( ostream& os = cout, bool pluginCall = false );
    
    // Normalization factor
    double getEvNormFac(int iAntPhys) {
      double evNorm = (iAntPhys<iQQemitIF ? evNormIIpT2 : evNormIFpT2);
      if (!evAllPT) {
        if (iAntPhys == iQXsplitII || iAntPhys == iQXsplitIF) evNorm = evNormMsplitI;
        else if (iAntPhys == iGXconvII || iAntPhys == iGXconvIF) evNorm = evNormMconvI;
        else if (iAntPhys == iXGsplitIF) evNorm = evNormMsplitF;
      }
      return evNorm;
    }

    // Evolution Windows: Phase space region boundaries
    int getRegion(double q, int iSys) {
      if (iSys == iSysEvSamples && needEvSamples && bornConfig) {
        for (int i=1; i<(int)regMinScalesNow.size(); i++)
          if (q < regMinScalesNow[i]) return i-1;
        return (int)regMinScalesNow.size()-1;
      } else {
        for (int i=1; i<(int)regMinScalesSav.size(); i++)
          if (q < regMinScalesSav[i]) return i-1;
        return (int)regMinScalesSav.size()-1;
      }
      /*if (iSys == iSysEvSamples && needEvSamples && bornConfig) {
        if (q < mc/5) return 0;
        else if (q < mc) return 1;
        else if (q < mb) return 2;
        else if (q < min(mtb,Qfac)) return 3;
        else if (q < min(max(mtb,Qfac),mt)) return 4;
        else {
          double scale3 = max(mt,Qfac);
          if (q < scale3) return 5;
          else return int( log(q/scale3)/log(6) + 6); 
        }
      } else {
        if (q < mc/5) return 0;
        else if (q < mc) return 1;
        else if (q < mb) return 2;
        else if (q < mtb) return 3;
        else if (q < mt) return 4;
        else return int( log(q/mt)/log(5) + 5);
      }*/
    }

    // Evolution Windows: Phase space region boundaries
    double getQmin(int iRegion, int iSys) {
      iRegion = max(0,iRegion);
      if (iSys == iSysEvSamples && needEvSamples && bornConfig) {
        iRegion = min(iRegion,(int)regMinScalesNow.size()-1);
        return regMinScalesNow[iRegion];
      } else {
        iRegion = min(iRegion,(int)regMinScalesSav.size()-1);
        return regMinScalesSav[iRegion];
      }
      /*if (iSys == iSysEvSamples && needEvSamples && bornConfig) {
        if (iRegion <= 0) return mc/16;
        else if (iRegion == 1) return mc/5;
        else if (iRegion == 2) return mc;
        else if (iRegion == 3) return mb;
        else if (iRegion == 4) return min(mtb,Qfac);
        else if (iRegion == 5) return min(max(mtb,Qfac),mt);
        else if (iRegion == 6) return max(mt,Qfac);
        else return pow(6,(double)iRegion-6.0)*max(mt,Qfac);
      } else {
        if (iRegion <= 0) return mc/16;
        else if (iRegion == 1) return mc/5;
        else if (iRegion == 2) return mc;
        else if (iRegion == 3) return mb;
        else if (iRegion == 4) return mtb;
        else if (iRegion == 5) return mt;
        else return pow(5,(double)iRegion-5.0)*mt;
      }*/
    }

    // Number of active flavors 
    int getNf(int iRegion, int iSys) {
      if (iSys == iSysEvSamples && needEvSamples && bornConfig) {
        double scaleMin = getQmin(iRegion, iSys);
        if (scaleMin >= mt) return 6;
        else if (scaleMin >= mb) return 5;
        else if (scaleMin >= mc) return 4;
        else return 3;
      } else {
        if (iRegion <= 1) return 3;
        else if (iRegion <= 2) return 4;
        else if (iRegion <= 4) return 5;
        else return 6;
      }
    }

    // AlphaS, MSbar and CMW
    AlphaStrong* alphaSptr;
    AlphaStrong* alphaSCMWptr;
    AlphaStrong* alphaSdefPtr;
    AlphaStrong* alphaSCMWdefPtr;

    // Lambda value
    double getLambda(int nFin) {
      if (nFin <= 3) return alphaSptr->Lambda3();
      else if (nFin <= 4) return alphaSptr->Lambda4();
      else if (nFin <= 5) return alphaSptr->Lambda5();
      else return alphaSptr->Lambda6();
    }
    double getLambdaCMW(int nFin) {
      if (nFin <= 3) return alphaSCMWptr->Lambda3();
      else if (nFin <= 4) return alphaSCMWptr->Lambda4();
      else if (nFin <= 5) return alphaSCMWptr->Lambda5();
      else return alphaSCMWptr->Lambda6();
    }

    // Functions to initialize, set and get headRoomFactors
    void initHeadRoom(double valEmit, double valSplitII, double valSplitIF, 
                      double valConvII, double valConvIF, double valSplitF) {
      for (int iTrialGen=0; iTrialGen<=13; iTrialGen++) {
        double valNow = 1.0;
        if ( iTrialGen<3 || (iTrialGen>4 && iTrialGen<11) ) valNow = valEmit;
        else if (iTrialGen == 3)  valNow = valSplitII;
        else if (iTrialGen == 11) valNow = valSplitIF;
        else if (iTrialGen == 4)  valNow = valConvII;
        else if (iTrialGen == 13) valNow = valConvIF;
        else if (iTrialGen == 12) valNow = valSplitF;
        // Loop over regions
        for (int iReg=0; iReg<=20; iReg++) {
          // Loop over number of gluons
          for (int nGin=0; nGin<=8; nGin++) {
            // Loop over number of quarks
            for (int nQin=0; nQin<=8; nQin++) {
              headRoom[iTrialGen][iReg][nGin][nQin]             = valNow;
              headRoomMECs[iTrialGen][iReg][nGin][nQin]         = 2.0*valNow;
              headRoomOptimize[iTrialGen][iReg][nGin][nQin]     = optimizeHeadRoom;
              headRoomMECsOptimize[iTrialGen][iReg][nGin][nQin] = optimizeHeadRoom;
            }
          }
        }
      }
    }
    double getHeadRoom(int iTrialGen, int iReg, int nGin, int nQin,
                       int iSys, double qMin) {
      if (!iRegDependance) iReg = 1;
      else if (iReg == -1) iReg = getRegion(qMin, iSys);
      if (iTrialGen < 0  || iReg < 0  || nGin < 0 || nGin > 8 || 
          iTrialGen > 13 || iReg > 20 || nQin < 0 || nQin > 8) return 1.0;
      if ((nBranchs[iSys]+1<=matchingLO) && hardSys[iSys])
        return headRoomMECs[iTrialGen][iReg][nGin][nQin];
      else return headRoom[iTrialGen][iReg][nGin][nQin];
    }
    void setHeadRoom(int iTrialGen, int iReg, int nGin, int nQin,
                     int iSys, double qMin, double scale) {
      if (!iRegDependance) iReg = 1;
      else if (iReg == -1) iReg = getRegion(qMin, iSys);
      if (iTrialGen < 0  || iReg < 0  || nGin < 0 || nGin > 8 || 
          iTrialGen > 13 || iReg > 20 || nQin < 0 || nQin > 8) return;
      bool MEC      = ((nBranchs[iSys]+1<=matchingLO) && hardSys[iSys]);
      double maxVal = 5.0;
      double oldVal = (MEC ? headRoomMECs[iTrialGen][iReg][nGin][nQin]
                           : headRoom[iTrialGen][iReg][nGin][nQin]);
      // Switch off further optimization if increasing headroom factor
      if ( scale > 1.0 ) {
        if (MEC) {
          headRoomMECsOptimize[iTrialGen][iReg][nGin][nQin] = false;
          headRoomMECs[iTrialGen][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
        } else {
          headRoomOptimize[iTrialGen][iReg][nGin][nQin] = false;
          headRoom[iTrialGen][iReg][nGin][nQin] = min(scale*oldVal,maxVal);
        }
        if (verbose >= 5) {
          printOut("SpaceAntShower::setHeadRoom","Increasing headroom at nG = "
                   +num2str(nGin,2)+" nQ = "+num2str(nQin,2)+"; current "
                   "headRoom = "+num2str(oldVal));
        }
      } else {
        // Allow reduction of headroom if we're still optimizing
        if (MEC) {
          if (headRoomMECsOptimize[iTrialGen][iReg][nGin][nQin])
            headRoomMECs[iTrialGen][iReg][nGin][nQin] = scale*oldVal;
        } else {
          if (headRoomOptimize[iTrialGen][iReg][nGin][nQin])
            headRoom[iTrialGen][iReg][nGin][nQin] = scale*oldVal;
        }
      }
    }

    // Add trial functions to a Branchelemental
    void resetTrialGenerators(BranchelementalISR* trial);

    // Method to check if a gluon splitting in the initial state (to get rid of
    // heavy quarks) is still possible after the current branching
    bool checkHeavyQuarkPhaseSpace(vector<Particle> parts, int iSyst);
    
    // Method to check if we have heavy quark left after passing the evolution window
    bool heavyQuarkLeft(double qTrial);
    
    // Function to set Markovian restart scale and who is unordered
    // and the particles in the system (to save time in update)
    void setMarkovianProps(int iSys, double Q2restart, vector<int> iUnOrd,
                           vector<Particle> particlesIn, 
                           vector<int> particleNrsIn) {
      if ((int)Q2hat.size() > iSys) {
        Q2hat[iSys]      = Q2restart;
        iUnOrdered[iSys] = iUnOrd;
      }
      if ((int)particleNrs.size() > iSys) {
        particleNrs[iSys] = particleNrsIn;
        particles[iSys]   = particlesIn;
      }
    }
    
    // Function to ask if a system is hard
    bool isSysHard(int iSys) {
      if (!isInit) return false;
      if ((int)hardSys.size() <= iSys) return false;
      return hardSys[iSys];
    }
    
    // Function to access pdfReWeight
    double PDFweight() {
      if (isInit) return pdfReWeight;
      return 1.0;
    }
    
    vector<double> getMasses() {
      vector<double> masses;
      masses.resize(0);
      masses.push_back(mt);
      masses.push_back(mtb);
      masses.push_back(mb);
      masses.push_back(mc);
      masses.push_back(ms);
      return masses;
    }
    pair<bool,bool> getSamplesInfo() {
      return make_pair(needEvSamples,bornConfig);
    }
    
    // Set verbose level
    void setVerbose(int verboseIn) { verbose = verboseIn; }

    
  private : 
    
    // -----------------------------------------------------------------------
    
    // PRIVATE MEMBERS
    
    // Generate trial branching kinematics and check physical phase space
    bool generateKinematics(BranchelementalISR* trialPtr, vector<Vec4>& pRec) {
      if (trialPtr->isII()) 
        return generateKinematicsII(trialPtr, pRec);
      else 
        return generateKinematicsIF(trialPtr);
    }
    bool generateKinematicsII(BranchelementalISR* trialPtr, vector<Vec4>& pRec);
    bool generateKinematicsIF(BranchelementalISR* trialPtr);
    
    // Main trial accept function
    bool acceptTrial(Event& event, BranchelementalISR* winnerPtr, 
                     vector<Particle> partsNow, double Q2hatNew);

    // Method to assign colour flow
    bool assignColourFlow(Event& event, BranchelementalISR* trialPtr);

    // Initialized?
    bool isInit;
    
    // Beams
    int beamFrameType;
    BeamParticle* beamAPtr;
    BeamParticle* beamBPtr;
    double eBeamA, eBeamB, shhSav;
    
    // Possibility to allow user veto of emission step.
    bool hasUserHooks, canVetoEmission;

    // Shower parameters
    int orderingMode, orderingModeSav, orderingModeMPI;
    int nGluonToQuarkI, nGluonToQuarkF;
    double cutoffScaleII, cutoffScaleIF;
    double evNormIIpT2, evNormIFpT2;
    double evNormMsplitI, evNormMconvI, evNormMsplitF;
    bool evAllPT;
    int nFlavZeroMass, massCorrectionLevel;
    bool treatAsMassless;
    bool doAcceptCombiII, doAcceptCombiIF;

    // AlphaS parameters
    int alphaSorder, CMWtypeII, CMWtypeIF; 
    double alphaSvalue, alphaSmax, alphaSmuFreeze, alphaSmuMin;
    double aSkMu2, aSkMu2SplitI, aSkMu2SplitF, aSkMu2Conv;
    double mt, mtb, ms, mb, mc;
    // Defaults
    int alphaSorderDef, CMWtypeIIDef, CMWtypeIFDef; 
    double alphaSvalueDef, alphaSmaxDef, alphaSmuFreezeDef, alphaSmuMinDef;
    double aSkMu2Def, aSkMu2SplitIDef, aSkMu2SplitFDef, aSkMu2ConvDef;
    // Calculated values
    double mu2freeze, mu2min;
    double mu2freezeDef, mu2minDef;
    
    // Trial Generators
    TrialIISoft    trialIISoft;
    TrialIIGCollA  trialIIGCollA;
    TrialIIGCollB  trialIIGCollB;
    TrialIISplitA  trialIISplitA;
    TrialIISplitB  trialIISplitB;
    TrialIIConvA   trialIIConvA;
    TrialIIConvB   trialIIConvB;
    TrialIFSoft    trialIFSoft;
    TrialIFQCollA  trialIFQCollA;
    TrialIFQCollK  trialIFQCollK;
    TrialIFGCollA1 trialIFGCollA1;
    TrialIFGCollA2 trialIFGCollA2;
    TrialIFGCollK  trialIFGCollK;
    TrialIFSplitA  trialIFSplitA;
    TrialIFSplitK  trialIFSplitK;
    TrialIFConvA   trialIFConvA;

    // Uncertainty Bands
    bool uncertaintyBands;
    double kMu2Var, uncertaintyBandsMaxDev;

    // Colour Reconnection parameters
    int CRmode;
    bool CRjunctions;
    
    // Desired matching multiplicities, level, regulator, regulator scales
    MEcorrections* MEcorrectionsPtr;
    bool meCorrections;
    int  matchingLO, matchingNLO;
    int  matchingRegOrder, matchingRegShape;
    double matchingScale, matchingScaleRatio, matchingIRcutoff;
    bool matchingScaleIsAbs, matchingFullColor;
    bool Hinterference;
    bool isMGOn; // True if real MG interface

    // Pointer to Evolution Variable
    Resolution* resolutionPtr;
   
    // Pointer to VINCIA common, functions, clustering
    VinciaCommon* vinComPtr;
    Functions* functionsPtr;
    VinClu* vinCluPtr;
    
    // Hooks for vetos
    VinciaVetoHooks* vetoHooksPtr;
    
    // Pointer to weights class
    VinciaWeights* weightsPtr;
    
    // Pointer to VINCIA FSR shower
    TimeAntShower* FSRshowerPtr;

    // Pointer to VINCIA Reconnections object
    Reconnections* reconnectionsPtr;

    // New headRoomFactors
    // first arg = iTrialGen, second arg = iRegion, third arg = nG, fourth arg = nG
    double headRoom[14][21][9][9];
    bool   headRoomOptimize[14][21][9][9];
    double headRoomMECs[14][21][9][9];
    bool   headRoomMECsOptimize[14][21][9][9];
    bool   optimizeHeadRoom, iRegDependance;
    
    // Vector of antenna sets (0 is physical, >= 1 are variations, not used so far)
    vector<AntennaSetISR*> antennaSets;
    
    // Total and MEC accept probability
    vector<double> Paccept;
    vector<double> PME;
    
    // Only relevant for the hard process
    // Are we a QCD process? Colour objects in the final state, unless they
    // they arise from a decay, i.e. gg -> H -> gg is NOT a QCD process
    bool isQCD;
    // Factorization scale
    double Q2fac, Qfac;
    // Are we doing allowing hard jets from first branching?
    bool hardJets, hardJetsQCD;
    int hardJetsAlphaS;
    // Enhancement factor for first branching
    double enhanceQ2fac;
    // Do we need different event samples? In which System?
    bool needEvSamples;
    int iSysEvSamples;
    // Did we decide to go for a Born or a Born+1 configuration?
    bool bornConfig;
    // Do we need to create the Delta0 in the Born sample?
    bool createDelta0;
    // Weight from using two event samples
    double pdfReWeight;
    // Vincia used for merging in Pythia?
    bool doMergeSav;
    
    // Evolution windows
    vector<double> regMinScalesMtSav;
    vector<double> regMinScalesSav;
    vector<double> regMinScalesNow;

    // Current winner
    BranchelementalISR* winnerPtr;
    int indxWin;
    int iSysWin;
    
    // ---------------------------- Every system gets own copy -------------------------------- //
    
    // Is system a hard system?
    vector<bool> hardSys;
    
    // Matrix element class for every system
    vector<int> MEclass;
    vector<int> nQLneeded;
    vector<int> Hint;

    // Saved particle state, nr in event record and particle
    vector< vector< int      > > particleNrs;
    vector< vector< Particle > > particles;

    // vector of dipoles (with trial branchings, 4 at most)
    vector< vector<BranchelementalISR> > branchelementals;
    
    // Markovian restart scale, min[scale of possible last branchings]
    vector<double> Q2hat;
    vector< vector<int> > iUnOrdered;
    
    // Count the number of branchings in the system
    vector<int> nBranchs;
    
    // Are we doing the first trial in the system
    vector<bool> firstTrial;
    
    // Saved incoming guys
    vector<Particle> initialA;
    vector<Particle> initialB;
    double eAused;
    double eBused;
    
    // Count numbers of quarks and gluons
    vector<int> nG, nQQ;
    
    // PDF freezing scale, the scale at which pdfs were evaluated last time
    // First entry: Scale of last evaluation
    // Second entry: Scale of current evaluation
    // Similar for the x-values
    vector<double> Q2freezeA;
    vector<double> Q2freezeB;
    
    // ---------------------------- //
    
    // Number of weights for uncertainty bands
    int nWeights;
    // Reweighting in case of MC violations
    bool allowReweighting;
    
    // Statistics
    int nTrialsSum, nTrialsVeto;
    int nTrialsHull, nTrialsHullMassive;
    int nTrialsCutM, nTrialsCutQ;
    int nBornVeto, nBornPlusJetVeto;
    int nClosePSforHQ, nForceSplit;
    int nMCviolShower, nMCviolDidMEC, nMCviolPlannedMEC;
    double maxMCviolShower, maxMCviolDidMEC, maxMCviolPlannedMEC;
    int nNegPaShower, nNegPaDidMEC, nNegPaPlannedMEC;
    double maxNegPaShower, maxNegPaDidMEC, maxNegPaPlannedMEC;
    
    // Rescue mechanism
    bool doRescue;
    int nRescue;
    double rescueMin;

    // Verbose setting, P>1 warning settings, and version number
    int verbose;
    double pAcceptFuss, pAcceptPT0;
    int nReportMCviol;
    double version;
    bool headerIsPrinted;
    
    // Counter for numbers of events
    long nAccepted, nSelected;
    int nVetoUserHooks, nFailHadLevel, nCallPythiaNext;
    int startDebug;
    
  };
  
}

#endif
