// MEcorrections.cc is a part of the VINCIA plug-in to the 
// PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "Vincia/MEcorrections.h"

namespace Vincia {
  
  //************************************************************************
  
  // Initialize
  
  void MEcorrections::init() {
    verbose            = settingsPtr->mode("Vincia:verbose");
    orderingMode       = settingsPtr->mode("Vincia:orderingMode");
    // opt 3 (smooth up to mLO) = opt 2 (smooth everywhere) here
    orderingMode       = min(orderingMode, 2);
    
    // For QCD processes
    hardJetsQCD        = settingsPtr->flag("Vincia:hardJetsQCD");
    factorScale2       = settingsPtr->mode("SigmaProcess:factorScale2");
    factorMultFac      = settingsPtr->parm("SigmaProcess:factorMultFac");
    factorFixScale     = settingsPtr->parm("SigmaProcess:factorFixScale");
    
    // Evolution types
    evolTypeGemiFF     = settingsPtr->mode("Vincia:evolutionType");
    evolTypeGsplitFF   = 7;
    evolTypeGemiI      = 1;
    evolTypeGsplitI    = (settingsPtr->flag("Vincia:evolveAllInPT") ? 1 : 2);
    evolTypeGsplitF    = (settingsPtr->flag("Vincia:evolveAllInPT") ? 1 : 4);
    evolTypeQconvI     = (settingsPtr->flag("Vincia:evolveAllInPT") ? 1 : 3);
    
    // user alphaS settings, valid for ISR and FSR
    aSmax              = settingsPtr->parm("Vincia:alphaSmax"); 
    aSmuFreeze         = settingsPtr->parm("Vincia:alphaSmuFreeze");
    mu2freeze          = pow2(aSmuFreeze);
    aSmuMin            = max( aSmuFreeze, 1.05 * max( aSptr->Lambda3(),
                                                      aSCMWptr->Lambda3() ) );
    mu2min             = pow2(aSmuMin);
    if (settingsPtr->mode("Vincia:alphaSorder") >= 1)
      aSmax = min( aSmax, min( aSptr->alphaS(mu2min),
                               aSCMWptr->alphaS(mu2min) ) );
      
    // default alphaS settings, valid for ISR and FSR
    aSmaxDef           = settingsPtr->parmDefault("Vincia:alphaSmax"); 
    aSmuFreezeDef      = settingsPtr->parmDefault("Vincia:alphaSmuFreeze");
    mu2freezeDef       = pow2(aSmuFreezeDef);
    aSmuMinDef         = max( aSmuFreezeDef, 1.05 * max( aSdefPtr->Lambda3(),
                                                         aSCMWdefPtr->Lambda3() ) );
    mu2minDef          = pow2(aSmuMinDef);
    if (settingsPtr->modeDefault("Vincia:alphaSorder") >= 1)
      aSmaxDef = min( aSmaxDef, min( aSdefPtr->alphaS(mu2minDef),
                                     aSCMWdefPtr->alphaS(mu2minDef) ) );
    
    // alphaS settings for FSR
    CMWtypeFF          = settingsPtr->mode("Vincia:CMWtypeFF");
    aSkMu2F            = pow2(settingsPtr->parm("Vincia:alphaSkMuF"));
    CMWtypeFFdef       = settingsPtr->modeDefault("Vincia:CMWtypeFF");
    aSkMu2Fdef         = pow2(settingsPtr->parmDefault("Vincia:alphaSkMuF"));
    
    // alphaS settings for ISR
    CMWtypeII          = settingsPtr->mode("Vincia:CMWtypeII");
    CMWtypeIF          = settingsPtr->mode("Vincia:CMWtypeIF");
    aSkMu2I            = pow2(settingsPtr->parm("Vincia:alphaSkMuI"));
    CMWtypeIIdef       = settingsPtr->modeDefault("Vincia:CMWtypeII");
    CMWtypeIFdef       = settingsPtr->modeDefault("Vincia:CMWtypeIF");
    aSkMu2Idef         = pow2(settingsPtr->parmDefault("Vincia:alphaSkMuI"));
    
    // Calculated values
    
    // Matching settings
    matchingLO         = settingsPtr->mode("Vincia:matchingLO");
    meCorrections      = (matchingLO >= 1);
    matchingFullColor  = settingsPtr->flag("Vincia:matchingFullColor");
    
    // Uncertainties
    uncertaintyBands   = settingsPtr->flag("Vincia:uncertaintyBands");
    nWeights           = (uncertaintyBands ? 11 : 1);
    aSkMuVar2          = pow2(settingsPtr->parm("Vincia:uncertaintyBandsKmu"));
    
    // Set initial values
    procClass          = 0;
    doPari             = false;
    nQLmin             = 0;
    doCMWnow           = true;
    bool isMGOn = settingsPtr->flag("Vincia:MGInterface");
    if (meCorrections && !isMGOn) {
      if (verbose >= 1) printOut("Vincia::MEcorrections","MGInterface set to off, "
                                 "but matchingLO >= 1. Switching it on now.");
      isMGOn = true;
      settingsPtr->flag("Vincia:MGInterface",isMGOn);
    }
    if (isMGOn) {
      isMGOn = mgInterfacePtr->init();
      mgInterfacePtr->setColorDepth(matchingFullColor);
      settingsPtr->flag("Vincia:MGInterface",isMGOn);
    }
    if (isMGOn && verbose >= 2) printOut("Vincia::MEcorrections", "MGInterface switched on.");
    else if (!isMGOn && verbose >= 1) printOut("Vincia::MEcorrections", "Couldn't switch MGInterface on.");
  
    isInit             = true;
  }
  
  //************************************************************************
  
  // Main Function: Get all possible histories and perform MEn+1 / Sum ant*MEn
  
  vector<double> MEcorrections::MEcorrection(vector<Particle> parts) {
    // Result
    vector<double> MEoverPS;
    MEoverPS.resize(nWeights);
    for (int i=0; i<(int)MEoverPS.size(); i++) MEoverPS[i] = 1.0;
    
    // Safety checks
    if (!canDoMEC(parts, false)) return MEoverPS;
    
    if (verbose >= 6) {
      printOut("Vincia::MEcorrections", "Printing post-branching momenta: id col acol p");
      for (int i=0; i<(int)parts.size(); i++)
        cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col()
             << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
    }
    
    // Check if we have the first emission in QCD events
    bool firstBranchQCD = (procClass == pureQCD && parts.size() == 5);
    
    // Mn+1
    double MEn1 = evalME(parts);
    
    // TODO MEn+1 < 0 should not happen
    if (MEn1 < 0.0) {
      if (verbose >= 3) {
        printOut("Vincia::MEcorrections",
                 "WARNING! ME(n+1) < 0 (" + num2str(MEn1) + "), "
                 " printing post-branching momenta: id col acol p");
        double pSum[4] = {0.0,0.0,0.0,0.0};
        for (int i=0; i<(int)parts.size(); i++) {
          cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col()
               << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
          if (parts[i].isFinal()) {
            pSum[0] += parts[i].px();
            pSum[1] += parts[i].py();
            pSum[2] += parts[i].pz();
            pSum[3] += parts[i].e();
          }
        }
        cout << "  " << setw(5) << "pSum:    ";
        for (int i=0; i<4; i++) cout << scientific << "  " << setw(5) << pSum[i];
        cout << endl;
      }
      return MEoverPS;
    }
    
    // Perform the clustering
    clustering(parts);
    // Copies
    vector< vector< vector<Particle> > > clusteredParts = clusteredPartsSav;
    vector< vector<        int       > > iAnts          = iAntSav;
    vector< vector<        double    > > scale2         = scale2Sav;
    vector< vector< vector<double  > > > invariants     = invariantsSav;
    
    vector<double> sum;
    sum.resize(nWeights);
    vector<double> addToMEn1;
    addToMEn1.resize(nWeights);
    for (int i=0; i<(int)sum.size(); i++) {
      sum[i] = 0.0;
      addToMEn1[i] = 0.0;
    }
    
    // Loop over clusterings
    for (int iState=0; iState<(int)clusteredParts.size(); iState++) if (iState<2) {
      for (int iCl=0; iCl<(int)clusteredParts[iState].size(); iCl++) {
        
        // TODO this is an ugly check for W production, don't know how to 
        // prevent the clustering from happening in the first place
        bool skipClustering = false;
        if (clusteredParts[iState][iCl].size() == 5) {
          bool foundW = false;
          for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++)
            if (clusteredParts[iState][iCl][iP].idAbs() == 24) foundW = true;
          if (foundW) {
            bool quarkOnly = true;
            for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++)
              if (clusteredParts[iState][iCl][iP].idAbs() == 21) quarkOnly = false;
            if (quarkOnly) {
              int flavSumIS = 0;
              int flavSumFS = 0;
              int iQI  = -1;
              int iQF  = -1;
              int iQBF = -1;
              for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++) {
                int idNow = clusteredParts[iState][iCl][iP].id();
                if (abs(idNow) < 10) {
                  if (clusteredParts[iState][iCl][iP].isFinal()) {
                    flavSumFS += idNow;
                    if (idNow > 0) iQF = iP;
                    else iQBF = iP;
                  }
                  else {
                    flavSumIS += idNow;
                    if (idNow > 0) iQI = iP;
                  }
                }
              }
              if (flavSumIS == 0 && flavSumFS != 0)
                skipClustering = !( (clusteredParts[iState][iCl][iQI].id() == 
                                     clusteredParts[iState][iCl][iQF].id()) ||
                                    (clusteredParts[iState][iCl][iQI].id() == 
                                     -clusteredParts[iState][iCl][iQBF].id()) );
            }
          }
        }
        if (skipClustering) continue;
        
        int iant    = iAnts[iState][iCl];
        double s1j  = invariants[iState][iCl][0];
        double sj2  = invariants[iState][iCl][1];
        double S12  = invariants[iState][iCl][3];
        vector<double> antenna;
        antenna.resize(nWeights); 
        
        //************************************************************************
        // iVar = 0 : User settings
        //        1 : Default settings
        //        2 : AlphaS-Hi (user settings, alphaS(Q/k))
        //        3 : AlphaS-Lo (user settings, alphaS(Q*k))
        //        4 : Ant-Hi    (user settings, MAX ants)
        //        5 : Ant-Lo    (user settings, MIN ants)
        //        6 : NLO-Hi    (user settings, *(1+alphaS))
        //        7 : NLO-Lo    (user settings, /(1+alphaS))
        //        8 : NLC-Hi    (user settings, CA for all emissions)
        //        9 : NLC-Lo    (user settings, 2CF for all emissions)
        //       10 : Pimp      (user settings, Pimp = 1/(1+q^4/Q^4)
        //************************************************************************
        
        // Add alpha(pT^2)/2pi * K * eik * M^2
        // to numerator if CMW rescaling is on
        vector<double> preFacCMW;
        
        if (iState == 0) { // ISR
          antenna[0] =  getAntISR(iant)->antennaFunction(s1j,sj2,S12)
                       *getAntISR(iant)->getChargeFactor();
          bool isEmi = ( iant <= iGGemitII || (iant >= iQQemitIF && iant <= iGGemitIF) );
          
          // CMW
          if (isEmi) {
            bool isII   = (iant <= iGGemitII);
            double eik1 = 2.0*S12/s1j/sj2;
            double eik2 = (isII ? 2.0*(S12+s1j+sj2)/s1j/sj2
                                : 2.0*(S12+sj2-s1j)/s1j/sj2);
            double mu2  = s1j*sj2 / (isII ? (S12+s1j+sj2) : (S12+sj2));
            // User
            bool CMWtypeUsr = (isII ? CMWtypeII : CMWtypeIF);
            // Special choice for first branching in hard jet event sample
            if (!doCMWnow) CMWtypeUsr = 0;
            double eikUsr   = 0.0;
            if (CMWtypeUsr > 0) eikUsr = (CMWtypeUsr == 1 ? eik1 : eik2);
            double mu2Usr     = max(mu2min, mu2freeze + aSkMu2I*mu2);
            double aSratioUsr = min(aSmax, aSCMWptr->alphaS(mu2Usr)) /
                                min(aSmax, aSptr->alphaS(mu2Usr));
            preFacCMW.push_back(eikUsr*(aSratioUsr-1.0));
            // alphaS variations
            if (uncertaintyBands) {
              // Default
              bool CMWtypeDef   = (isII ? CMWtypeIIdef : CMWtypeIFdef);
              // Special choice for first branching in hard jet event sample
              if (!doCMWnow) CMWtypeDef = 0;
              double eikDef     = 0.0;
              if (CMWtypeDef > 0) eikDef = (CMWtypeDef == 1 ? eik1 : eik2);
              double mu2Def     = max(mu2minDef, mu2freezeDef + aSkMu2Idef*mu2);
              double aSratioDef = min(aSmaxDef, aSCMWdefPtr->alphaS(mu2Def)) /
                                  min(aSmaxDef, aSdefPtr->alphaS(mu2Def));
              preFacCMW.push_back(eikDef*(aSratioDef-1.0)); // iVar = 1
              // Additional scaling factor
              double mu2V2      = max(mu2min, mu2freeze + aSkMu2I*mu2/aSkMuVar2);
              double aSratioV2  = min(aSmax, aSCMWptr->alphaS(mu2V2)) /
                                  min(aSmax, aSptr->alphaS(mu2V2));
              preFacCMW.push_back(eikUsr*(aSratioV2-1.0)); // iVar = 2
              double mu2V3      = max(mu2min, mu2freeze + aSkMu2I*mu2*aSkMuVar2);
              double aSratioV3  = min(aSmax, aSCMWptr->alphaS(mu2V3)) /
                                  min(aSmax, aSptr->alphaS(mu2V3));
              preFacCMW.push_back(eikUsr*(aSratioV3-1.0)); // iVar = 3
              // iVar = 4..10
              for (int iVar=4; iVar<nWeights; ++iVar)
                preFacCMW.push_back(eikUsr*(aSratioUsr-1.0));
            }
          } else for (int iVar=0; iVar<nWeights; ++iVar) preFacCMW.push_back(0.0);
          
          // ant variations
          if (uncertaintyBands) {
            for (int iVar=1; iVar<nWeights; ++iVar) {
                   if (iVar==1) antenna[iVar] =  getAntISR(iant,1)->antennaFunction(s1j,sj2,S12)
                                                *getAntISR(iant,1)->getChargeFactor();
              else if (iVar==4) antenna[iVar] =  getAntISR(iant,2)->antennaFunction(s1j,sj2,S12)
                                                *getAntISR(iant,2)->getChargeFactor();
              else if (iVar==5) antenna[iVar] =  getAntISR(iant,3)->antennaFunction(s1j,sj2,S12)
                                                *getAntISR(iant,3)->getChargeFactor();
              else if (iVar==8 && isEmi) 
                antenna[iVar] = getAntISR(iant)->antennaFunction(s1j,sj2,S12) *CA;
              else if (iVar==9 && isEmi) 
                antenna[iVar] = getAntISR(iant)->antennaFunction(s1j,sj2,S12) *CF;
              else antenna[iVar] =  getAntISR(iant)->antennaFunction(s1j,sj2,S12)
                                   *getAntISR(iant)->getChargeFactor();
            }
          }
        } 
        
        else { // FSR
          antenna[0] =  getAntFSR(iant)->antennaFunction(s1j/S12,sj2/S12)/S12
                       *getAntFSR(iant)->getChargeFactor();
          bool isEmi = ( iant <= iGGemit );
          
          // CMW
          if (isEmi) {
            double eik1 = 2.0*S12/s1j/sj2;
            double eik2 = 2.0*(S12-s1j-sj2)/s1j/sj2;
            double mu2  = s1j*sj2/S12;
            // User
            double eikUsr   = 0.0;
            if (CMWtypeFF > 0) eikUsr = (CMWtypeFF == 1 ? eik1 : eik2);
            double mu2Usr     = max(mu2min, mu2freeze + aSkMu2F*mu2);
            double aSratioUsr = min(aSmax, aSCMWptr->alphaS(mu2Usr)) /
                                min(aSmax, aSptr->alphaS(mu2Usr));
            preFacCMW.push_back(eikUsr*(aSratioUsr-1.0));
            // alphaS variations
            if (uncertaintyBands) {
              // Default
              double eikDef     = 0.0;
              if (CMWtypeFFdef > 0) eikDef = (CMWtypeFFdef == 1 ? eik1 : eik2);
              double mu2Def     = max(mu2minDef, mu2freezeDef + aSkMu2Fdef*mu2);
              double aSratioDef = min(aSmaxDef, aSCMWdefPtr->alphaS(mu2Def)) /
                                  min(aSmaxDef, aSdefPtr->alphaS(mu2Def));
              preFacCMW.push_back(eikDef*(aSratioDef-1.0)); // iVar = 1
              // Additional scaling factor
              double mu2V2      = max(mu2min, mu2freeze + aSkMu2F*mu2/aSkMuVar2);
              double aSratioV2  = min(aSmax, aSCMWptr->alphaS(mu2V2)) /
                                  min(aSmax, aSptr->alphaS(mu2V2));
              preFacCMW.push_back(eikUsr*(aSratioV2-1.0)); // iVar = 2
              double mu2V3      = max(mu2min, mu2freeze + aSkMu2F*mu2*aSkMuVar2);
              double aSratioV3  = min(aSmax, aSCMWptr->alphaS(mu2V3)) /
                                  min(aSmax, aSptr->alphaS(mu2V3));
              preFacCMW.push_back(eikUsr*(aSratioV3-1.0)); // iVar = 3
              // iVar = 4..10
              for (int iVar=4; iVar<nWeights; ++iVar)
                preFacCMW.push_back(eikUsr*(aSratioUsr-1.0));
            }
          } else for (int iVar=0; iVar<nWeights; ++iVar) preFacCMW.push_back(0.0);
          
          // ant variations
          if (uncertaintyBands) {
            for (int iVar=1; iVar<nWeights; ++iVar) {
                   if (iVar==1) antenna[iVar] =  getAntFSR(iant,1)->antennaFunction(s1j/S12,sj2/S12)/S12
                                                *getAntFSR(iant,1)->getChargeFactor();
              else if (iVar==4) antenna[iVar] =  getAntFSR(iant,2)->antennaFunction(s1j/S12,sj2/S12)/S12
                                                *getAntFSR(iant,2)->getChargeFactor();
              else if (iVar==5) antenna[iVar] =  getAntFSR(iant,3)->antennaFunction(s1j/S12,sj2/S12)/S12
                                                *getAntFSR(iant,3)->getChargeFactor();
              else if (iVar==8 && isEmi) 
                antenna[iVar] =  getAntFSR(iant)->antennaFunction(s1j/S12,sj2/S12)/S12 *CA;
              else if (iVar==9 && isEmi) 
                antenna[iVar] =  getAntFSR(iant)->antennaFunction(s1j/S12,sj2/S12)/S12 *CF;
              else antenna[iVar] =  getAntFSR(iant)->antennaFunction(s1j/S12,sj2/S12)/S12
                                   *getAntFSR(iant)->getChargeFactor();
            }
          }
        }
      
        // Ordering with Pimp factor
        double Pimp  = 1.0;
        double Pimp4 = 1.0;
        double Q2   = scale2[iState][iCl];
        // Calculate Markovian ordering scale
        double Q2hat = 0.0;
        vector<int> iUnOrd;
        // For first emission in QCD with hard jets allowed Q2hat is Q2fac
        if (firstBranchQCD) {
          if (factorScale2 == 5) Q2hat = factorFixScale;
          else if (factorScale2 == 4) { // sHat
            int iIS1 = -1;
            int iIS2 = -1;
            for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++) {
              if (!clusteredParts[iState][iCl][iP].isFinal()) {
                if (iIS1 == -1) iIS1 = iP;
                else if (iIS2 == -1) iIS2 = iP;
              }
            }
            Q2hat = factorMultFac*(clusteredParts[iState][iCl][iIS1].p()+
                                   clusteredParts[iState][iCl][iIS2].p()).m2Calc();
          } else { // pT2, no masses -> 1-3 = same result
            for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++) {
              if (clusteredParts[iState][iCl][iP].isFinal()) {
                Q2hat = factorMultFac*clusteredParts[iState][iCl][iP].pT2();
                break;
              }
            }
          }
        } else unorderingScale(clusteredParts[iState][iCl], &Q2hat, &iUnOrd);
        if (Q2hat > 0.0) { // 0 for first branching
          // Smooth ordering
          // (first branch in QCD allowed to be hard goes here)
          if (orderingMode == 2 || (firstBranchQCD && hardJetsQCD)) {
            Pimp  = 1.0/(1.0+Q2/Q2hat);
            Pimp4 = 1.0/(1.0+pow2(Q2/Q2hat));
            // Check that at least one of parents (0,1) appears in iUnOrd
            if (!(firstBranchQCD && hardJetsQCD)) {
              bool contrSmooth = (Q2hat > Q2);
              for (int i=0; i<(int)iUnOrd.size(); i++) {
                if (iUnOrd[i] == 0) contrSmooth = true;
                else if (iUnOrd[i] == 1) contrSmooth = true;
              }
              if (!contrSmooth) Pimp = 0.0;
            }
          }
          // Strong ordering
          // (first branch in QCD not hard goes here)
          else if (orderingMode == 1 || firstBranchQCD) if (Q2hat < Q2) Pimp = 0.0;
          if (Pimp == 0.0) Pimp4 = 0.0;
        }
        
        // Ariadne factor for final state gluon splitting
        double Pari = 1.0;
        if (doPari) if ( (iState == 0 && iant == iXGsplitIF) || 
                         (iState == 1 && iant == iQGsplit) || 
                         (iState == 1 && iant == iGGsplit) ) {
          int iGluon = 1; 
          int colG   = clusteredParts[iState][iCl][iGluon].col();
          int acolG  = clusteredParts[iState][iCl][iGluon].acol();
          int iN = -1;
          for (int i=2; i<(int)clusteredParts[iState][iCl].size(); i++) { // Find colour partner
            // Start at 2 to leave out parents
            int colN  = clusteredParts[iState][iCl][i].col();
            int acolN = clusteredParts[iState][iCl][i].acol();
            if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
          }
          double sN = 0.0;
          if (iN >= 0) {
            sN = m2(clusteredParts[iState][iCl][iGluon],clusteredParts[iState][iCl][iN]);
            Pari = 2.0*sN/(S12 + sN);      
          }
        }
    
        if (verbose >= 6) {
          printOut("Vincia::MEcorrections", 
                   "iAnt = " + num2str(iant) +
                   " Q2 = " + num2str(Q2) +
                   " Q2hat = " + num2str(Q2hat) +
                   " Pimp = " + num2str(Pimp) +
                   " Pari = " + num2str(Pari) +
                   " (orderingMode=" + num2str(orderingMode) + ")");
          cout << "  Printing pre-branching momenta (" << (iState == 0 ? "ISR" : "FSR") 
               << "): id col acol p:" << endl;
          for (int i=0; i<(int)clusteredParts[iState][iCl].size(); i++)
            cout << "  " << setw(2) << clusteredParts[iState][iCl][i].id()
                 << " " << setw(3) << clusteredParts[iState][iCl][i].col()
                 << " " << setw(3) << clusteredParts[iState][iCl][i].acol()
                 << " " << clusteredParts[iState][iCl][i].p();
        }
        
        // MEn
        double MEn = evalME(clusteredParts[iState][iCl]);
        
        // TODO MEn < 0 should not happen
        if (MEn < 0.0) {
          if (verbose >= 3) {
            printOut("Vincia::MEcorrections",
                     "WARNING! ME(n) < 0 (" + num2str(MEn) + "), "
                     " printing pre-branching momenta: id col acol p");
            double pSum[4] = {0.0,0.0,0.0,0.0};
            for (int i=0; i<(int)clusteredParts[iState][iCl].size(); i++) {
              cout << "  " << setw(2) << clusteredParts[iState][iCl][i].id() 
                   << " " << setw(3) << clusteredParts[iState][iCl][i].col()
                   << " " << setw(3) << clusteredParts[iState][iCl][i].acol() 
                   << " " << clusteredParts[iState][iCl][i].p();
              if (clusteredParts[iState][iCl][i].isFinal()) {
                pSum[0] += clusteredParts[iState][iCl][i].px();
                pSum[1] += clusteredParts[iState][iCl][i].py();
                pSum[2] += clusteredParts[iState][iCl][i].pz();
                pSum[3] += clusteredParts[iState][iCl][i].e();
              }
            }
            cout << "  " << setw(5) << "pSum:    ";
            for (int i=0; i<4; i++) cout << scientific << "  " << setw(5) << pSum[i];
            cout << endl;
          }
        }
        
        // Add CMW contribution
        // TODO do we want to have the Pimp factor here?
        addToMEn1[0] += preFacCMW[0]*MEn;
        if (uncertaintyBands) {
          for (int iVar=1; iVar<nWeights; ++iVar)
            addToMEn1[iVar] += preFacCMW[iVar]*MEn;
        }
    
        if (verbose >= 6) {
          cout << scientific;
          cout << "  antenna = " << antenna[0] << " MEn = " << MEn
               << " leadsto Pimp*Pari*antenna*MEn = " << Pimp*Pari*antenna[0]*MEn << endl;
          if (preFacCMW[0] > 0.0) 
            cout << "  CMW contribution to numerator = " << (preFacCMW[0]*MEn) << endl;
        }
        
        // Sum up
        sum[0] += Pimp*Pari*antenna[0]*MEn;
        if (uncertaintyBands) {
          for (int iVar=1; iVar<nWeights; ++iVar) {
            if (iVar==10) sum[iVar] += Pimp4*Pari*antenna[iVar]*MEn;
            else          sum[iVar] += Pimp *Pari*antenna[iVar]*MEn;
          }
        }
      }
    }
    
    if (verbose >= 6) printOut("Vincia::MEcorrections", "ME = " + num2str(MEn1) +
                               " PS = " + num2str(sum[0]) +
                               " leadsto ME/PS = " + num2str((MEn1+addToMEn1[0])/sum[0]) );
    
    MEoverPS[0] = (MEn1+addToMEn1[0])/sum[0];
    if (uncertaintyBands) {
      for (int iVar=1; iVar<nWeights; ++iVar) 
        MEoverPS[iVar] = (MEn1+addToMEn1[iVar])/sum[iVar];
    }
    
    if (sum[0] == 0.0) {
      if (verbose >= 2) {
        printOut("Vincia::MEcorrections", "sum over shower histories = 0!");
        if (firstBranchQCD) cout << " First branching in QCD event.";
        if (clusteredParts.size() == 0) cout << " Did not find any histories.";
        else cout << " Probably to due numerical instability.";
        cout << endl << " Putting PME to unity." << endl;
      }
      for (int iVar=0; iVar<nWeights; ++iVar) MEoverPS[iVar] = 1.0;
    }
                                                                                  
    // Done
    return MEoverPS;
  }
  
  //************************************************************************
  
  // Functions for flat phase space scan - first order
  
  void MEcorrections::PSME1stOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                                       vector<int>& nrOrdPaths, bool print, 
                                       bool isQCD) {
    verbose = (print ? 9 : 1);
    PSoverME.clear();   PSoverME.resize(3);   // 0 strong ordering
    nrOrdPaths.clear(); nrOrdPaths.resize(3); // 1 smooth ordering
    for (int i=0; i<3; i++) {                 // 2 no ordering
      PSoverME[i] = 0.0;
      nrOrdPaths[i] = 0;
    }
    
    if (print) {
      printOut("Vincia::MEcorrections::PSME1stOrderFlat", "Printing post-branching momenta: id col acol p");
      for (int i=0; i<(int)parts.size(); i++) {
        cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col();
        cout << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
      }
    }
    
    // Mn+1
    double MEn1 = evalME(parts);
    // Perform the clustering
    clustering(parts);
    // Copies
    vector< vector< vector<Particle> > > clusteredParts = clusteredPartsSav;
    vector< vector<        int       > > iAnts          = iAntSav;
    vector< vector<        double    > > scale2         = scale2Sav;
    vector< vector< vector<double  > > > invariants     = invariantsSav;
    
    // Loop over clusterings
    for (int iState=0; iState<(int)clusteredParts.size(); iState++) if (iState<2) {
      for (int iCl=0; iCl<(int)clusteredParts[iState].size(); iCl++) {
        int iant   = iAnts[iState][iCl];
        double s1j = invariants[iState][iCl][0];
        double sj2 = invariants[iState][iCl][1];
        double S12 = invariants[iState][iCl][3];
        double antenna   = 0.0;
        double preFacCMW = 0.0;
        if (iState == 0) { // ISR
          antenna  = getAntISR(iant)->antennaFunction(s1j,sj2,S12);
          antenna *= getAntISR(iant)->getChargeFactor();
          double eik = 0.0;
          double mu2 = 0.0;
          if ( iant <= iGGemitII && CMWtypeII > 0 ) {
            eik = (CMWtypeII == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12+s1j+sj2)/s1j/sj2);
            mu2 = aSkMu2I*s1j*sj2/(S12+s1j+sj2);
          } else if ( iant >= iQQemitIF && iant <= iGGemitIF && CMWtypeIF > 0 ) {
            eik = (CMWtypeIF == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12+sj2-s1j)/s1j/sj2);
            mu2 = aSkMu2I*s1j*sj2/(S12+sj2);
          }
          double mu2def  = max(mu2min, mu2freeze + mu2);
          double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                           min(aSmax, aSptr->alphaS(mu2def));
          preFacCMW = eik*(coupFac-1.0);
        } else { // FSR
          antenna  = getAntFSR(iant)->antennaMassless(s1j/S12,sj2/S12)/S12;
          antenna *= getAntFSR(iant)->getChargeFactor();
          if ( iant <= iGGemit && CMWtypeFF > 0 ) {
            double eik     = (CMWtypeFF == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12-sj2-s1j)/s1j/sj2);
            double mu2     = aSkMu2F*s1j*sj2/S12;
            double mu2def  = max(mu2min, mu2freeze + mu2);
            double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                             min(aSmax, aSptr->alphaS(mu2def));
            preFacCMW = eik*(coupFac-1.0);
          }
        }
        
        // For QCD check ordering wrt pT of hard process
        double Q2    = scale2[iState][iCl];
        double Q2fac = 0.0;
        bool ordered = true;
        double Pimp  = 1.0;
        if (isQCD) {
          for (int iP=0; iP<(int)clusteredParts[iState][iCl].size(); iP++) {
            if (clusteredParts[iState][iCl][iP].isFinal()) {
              Q2fac         = clusteredParts[iState][iCl][iP].pT2();
              ordered       = (Q2fac > Q2);
              break;
            }
          }
          Pimp = 1.0/(1.0+Q2/Q2fac);
        }
        
        // MEn
        double MEn = evalME(clusteredParts[iState][iCl]);
        // 0 strong ordering
        if (ordered) {
          PSoverME[0] += ( antenna * MEn );
          nrOrdPaths[0]++;
        }
        // 1 smooth ordering wrt pT of hard process
        PSoverME[1] += ( Pimp * antenna * MEn );
        nrOrdPaths[1]++;
        // 2 no ordering
        PSoverME[2] += ( antenna * MEn );
        nrOrdPaths[2]++;
        // all the same for non QCD
        
        // Add CMW contribution
        // TODO do we want to have the Pimp factor here?
        MEn1 += (preFacCMW*MEn);
        
        if (print) {
          printOut("Vincia::PSME1stOrderFlat", 
                   "iAnt = " + num2str(iant) +
                   " (orderingMode=" + num2str(orderingMode) + ")");
          cout << "  Printing pre-branching momenta (" << (iState == 0 ? "ISR" : "FSR") 
               << "): id col acol p:" << endl;
          for (int i=0; i<(int)clusteredParts[iState][iCl].size(); i++) {
            cout << "  " << setw(2) << clusteredParts[iState][iCl][i].id();
            cout << " " << setw(3) << clusteredParts[iState][iCl][i].col();
            cout << " " << setw(3) << clusteredParts[iState][iCl][i].acol();
            cout << " " << clusteredParts[iState][iCl][i].p();
          }
          cout << scientific;
          cout << "  antenna = " << antenna << " MEn = " << MEn;
          cout << " leadsto antenna*MEn = " << antenna*MEn << endl;
          if (preFacCMW > 0.0) 
            cout << "  CMW contribution to numerator = " << (preFacCMW*MEn) << endl;
        }
      }
    }
    
    for (int i=0; i<(int)PSoverME.size(); i++) 
      PSoverME[i] /= MEn1;
    
    if (print) printOut("Vincia::MEcorrections::PSME1stOrderFlat", "ME = " + num2str(MEn1));
  }
  
  // Functions for flat phase space scan - second order
  
  void MEcorrections::PSME2ndOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                                       vector< vector<double> >& Q2clust1, vector< vector<double> >& Q2clust2,
                                       vector<int>& nrOrdPaths, bool print, 
                                       bool isQCD, vector< vector< vector<double> > >& Q2clust2QCD,
                                       vector< vector< vector<double> > >& Q2fac) {
    verbose = (print ? 9 : 1);
    PSoverME.clear();    PSoverME.resize(3); // 0 strong ordering
    Q2clust1.clear();    Q2clust1.resize(3); // 1 smooth ordering
    Q2clust2.clear();    Q2clust2.resize(3); // 2 no ordering
    nrOrdPaths.clear();  nrOrdPaths.resize(3);
    Q2clust2QCD.clear(); Q2clust2QCD.resize(3);
    Q2fac.clear();       Q2fac.resize(3);
    for (int i=0; i<3; i++) {
      PSoverME[i] = 0.0;
      Q2clust1[i].clear(); Q2clust1[i].resize(0);
      Q2clust2[i].clear(); Q2clust2[i].resize(0);
      nrOrdPaths[i] = 0;
      Q2clust2QCD[i].clear(); Q2clust2QCD[i].resize(0);
      Q2fac[i].clear();       Q2fac[i].resize(0);
    }
    
    if (print) {
      printOut("Vincia::MEcorrections::PSME2ndOrderFlat", "Printing post-branching momenta: id col acol p");
      for (int i=0; i<(int)parts.size(); i++) {
        cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col();
        cout << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
      }
    }
    
    // Mn+1
    double MEn1 = evalME(parts);
    // Perform the clustering
    clustering(parts);
    // Copies
    vector< vector< vector<Particle> > > clusteredParts = clusteredPartsSav;
    vector< vector<        int       > > iAnts          = iAntSav;
    vector< vector<        double    > > scale2         = scale2Sav;
    vector< vector< vector<double  > > > invariants     = invariantsSav;
    
    // Loop over clusterings
    for (int iState=0; iState<(int)clusteredParts.size(); iState++) if (iState<2) {
      for (int iCl=0; iCl<(int)clusteredParts[iState].size(); iCl++) {
        int iant   = iAnts[iState][iCl];
        double s1j = invariants[iState][iCl][0];
        double sj2 = invariants[iState][iCl][1];
        double S12 = invariants[iState][iCl][3];
        double antenna = 0.0;
        double preFacCMW = 0.0;
        if (iState == 0) { // ISR
          antenna  = getAntISR(iant)->antennaFunction(s1j,sj2,S12);
          antenna *= getAntISR(iant)->getChargeFactor();
          double eik = 0.0;
          double mu2 = 0.0;
          if ( iant <= iGGemitII && CMWtypeII > 0 ) {
            eik = (CMWtypeII == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12+s1j+sj2)/s1j/sj2);
            mu2 = aSkMu2I*s1j*sj2/(S12+s1j+sj2);
          } else if ( iant >= iQQemitIF && iant <= iGGemitIF && CMWtypeIF > 0 ) {
            eik = (CMWtypeIF == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12+sj2-s1j)/s1j/sj2);
            mu2 = aSkMu2I*s1j*sj2/(S12+sj2);
          }
          double mu2def  = max(mu2min, mu2freeze + mu2);
          double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                           min(aSmax, aSptr->alphaS(mu2def));
          preFacCMW = eik*(coupFac-1.0);
        } else { // FSR
          antenna  = getAntFSR(iant)->antennaMassless(s1j/S12,sj2/S12)/S12;
          antenna *= getAntFSR(iant)->getChargeFactor();
          if ( iant <= iGGemit && CMWtypeFF > 0 ) {
            double eik     = (CMWtypeFF == 1 ? 2.0*S12/s1j/sj2 : 2.0*(S12-sj2-s1j)/s1j/sj2);
            double mu2     = aSkMu2F*s1j*sj2/S12;
            double mu2def  = max(mu2min, mu2freeze + mu2);
            double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                             min(aSmax, aSptr->alphaS(mu2def));
            preFacCMW = eik*(coupFac-1.0);
          }
        }
      
        // Ordering with Pimp factor
        double Pimp = 1.0;
        double Q2   = scale2[iState][iCl];
        // Calculate Markovian ordering scale
        double Q2hat = 0.0;
        vector<int> iUnOrd;
        unorderingScale(clusteredParts[iState][iCl], &Q2hat, &iUnOrd);
        // Everyone is unordered anyway since only two emissions
        Pimp = 1.0/(1.0+Q2/Q2hat);
        
        // Ariadne factor for final state gluon splitting
        double Pari = 1.0;
        if (doPari) if ( (iState == 0 && iant == iXGsplitIF) || 
                         (iState == 1 && iant == iQGsplit) || 
                         (iState == 1 && iant == iGGsplit) ) {
          int iGluon = 1; 
          int colG   = clusteredParts[iState][iCl][iGluon].col();
          int acolG  = clusteredParts[iState][iCl][iGluon].acol();
          int iN = -1;
          for (int i=2; i<(int)clusteredParts[iState][iCl].size(); i++) { // Find colour partner
            // Start at 2 to leave out parents
            int colN  = clusteredParts[iState][iCl][i].col();
            int acolN = clusteredParts[iState][iCl][i].acol();
            if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
          }
          double sN = 0.0;
          if (iN >= 0) {
            sN = m2(clusteredParts[iState][iCl][iGluon],clusteredParts[iState][iCl][iN]);
            Pari = 2.0*sN/(S12 + sN);      
          }
        }
        
        // MEn
        double MEn = evalME(clusteredParts[iState][iCl]);
        if (Q2hat > 0.0) { // should always be the case here
          // 0 strong ordering
          Q2clust1[0].push_back(Q2);
          Q2clust2[0].push_back(Q2hat);
          if (Q2hat > Q2) {
            PSoverME[0] += ( Pari * antenna * MEn );
            nrOrdPaths[0]++;
          }
          // 1 smooth ordering
          Q2clust1[1].push_back(Q2);
          Q2clust2[1].push_back(Q2hat);
          if (!isQCD) { // non QCD, everyone is unordered since first emission decides
            PSoverME[1] += ( Pimp * Pari * antenna * MEn );
            nrOrdPaths[1]++;
          } else { // QCD, not everyone is unordered
            bool contrSmooth = (Q2hat > Q2);
            for (int i=0; i<(int)iUnOrd.size(); i++) {
              if (iUnOrd[i] == 0) contrSmooth = true;
              else if (iUnOrd[i] == 1) contrSmooth = true;
            }
            if (contrSmooth) {
              PSoverME[1] += ( Pimp * Pari * antenna * MEn );
              nrOrdPaths[1]++;
            }
          }
          // 2 no ordering
          Q2clust1[2].push_back(Q2);
          Q2clust2[2].push_back(Q2hat);
          PSoverME[2] += ( Pari * antenna * MEn );
          nrOrdPaths[2]++;
        }
        
        // Add CMW contribution
        // TODO do we want to have the Pimp factor here?
        MEn1 += (preFacCMW*MEn);
        
        // For QCD we need the actual second clustering and factorization scale
        if (isQCD) {
          clustering(clusteredParts[iState][iCl]);
          // Copies
          vector< vector< vector<Particle> > > clusteredParts2 = clusteredPartsSav;
          vector< vector<        double    > > scale22         = scale2Sav;
          vector<double> secondClustScales;
          vector<double> facScales;
          for (int iState2=0; iState2<(int)clusteredParts2.size(); iState2++) if (iState2<2) {
            for (int iCl2=0; iCl2<(int)clusteredParts2[iState2].size(); iCl2++) { 
              secondClustScales.push_back(scale22[iState2][iCl2]);
              for (int iP=0; iP<(int)clusteredParts2[iState2][iCl2].size(); iP++) {
                if (clusteredParts2[iState2][iCl2][iP].isFinal()) {
                  facScales.push_back( clusteredParts2[iState2][iCl2][iP].pT2() );
                  break;
                }
              }
            }
          }
          for (int i=0; i<3; i++) {
            Q2clust2QCD[i].push_back(secondClustScales);
            Q2fac[i].push_back(facScales);
          }
        }
    
        if (print) {
          printOut("Vincia::MEcorrections::PSME2ndOrderFlat", 
                   "iAnt = " + num2str(iant) +
                   " Q2 = " + num2str(Q2) +
                   " Q2hat = " + num2str(Q2hat) +
                   " Pimp = " + num2str(Pimp) +
                   " Pari = " + num2str(Pari) +
                   " (orderingMode=" + num2str(2) + ")");
          cout << "  Printing pre-branching momenta (" << (iState == 0 ? "ISR" : "FSR") 
               << "): id col acol p:" << endl;
          for (int i=0; i<(int)clusteredParts[iState][iCl].size(); i++) {
            cout << "  " << setw(2) << clusteredParts[iState][iCl][i].id();
            cout << " " << setw(3) << clusteredParts[iState][iCl][i].col();
            cout << " " << setw(3) << clusteredParts[iState][iCl][i].acol();
            cout << " " << clusteredParts[iState][iCl][i].p();
          }
          cout << scientific;
          cout << "  antenna = " << antenna << " MEn = " << MEn;
          cout << " leadsto Pimp*Pari*antenna*MEn = " << Pimp*Pari*antenna*MEn << endl;
          if (preFacCMW > 0.0) 
            cout << "  CMW contribution to numerator = " << (preFacCMW*MEn) << endl;
        }
      }
    }
    
    for (int i=0; i<(int)PSoverME.size(); i++) 
      PSoverME[i] /= MEn1;
    
    if (print) printOut("Vincia::MEcorrections::PSME2ndOrderFlat", "ME = " + num2str(MEn1));
  }
  
  // Functions for flat phase space scan - third order
  
  void MEcorrections::PSME3rdOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                                       vector< vector<double> >& Q2clust1,
                                       vector<int>& nrOrdPaths, bool print) {
    verbose = (print ? 9 : 1);
    PSoverME.clear();   PSoverME.resize(6);   // 0 strong ordering,
    Q2clust1.clear();   Q2clust1.resize(6);   // 1 smooth ordering,
    nrOrdPaths.clear(); nrOrdPaths.resize(6); // 2 no ordering
    for (int i=0; i<6; i++) {
      PSoverME[i] = 0.0;
      nrOrdPaths[i] = 0;
    }
    
    if (print) {
      printOut("Vincia::MEcorrections::PSME3rdOrderFlat", "Printing post-branching momenta: id col acol p");
      for (int i=0; i<(int)parts.size(); i++) {
        cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col();
        cout << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
      }
    }
    
    // Mn+2
    double MEn2 = evalME(parts);
    // Perform the clustering
    clustering(parts);
    // Copies
    vector< vector< vector<Particle> > > clusteredParts1 = clusteredPartsSav;
    vector< vector<        int       > > iAnts1          = iAntSav;
    vector< vector<        double    > > scale21         = scale2Sav;
    vector< vector< vector<double  > > > invariants1     = invariantsSav;
    
    // Loop over first clusterings
    for (int iState1=0; iState1<(int)clusteredParts1.size(); iState1++) if (iState1<2) {
      for (int iCl1=0; iCl1<(int)clusteredParts1[iState1].size(); iCl1++) {
        int iant1   = iAnts1[iState1][iCl1];
        double s1j1 = invariants1[iState1][iCl1][0];
        double sj21 = invariants1[iState1][iCl1][1];
        double S121 = invariants1[iState1][iCl1][3];
        double antenna1  = 0.0;
        double preFacCMW = 0.0; // NOTE only sensible if matched to previous order
        if (iState1 == 0) { // ISR
          antenna1  = getAntISR(iant1)->antennaFunction(s1j1,sj21,S121);
          antenna1 *= getAntISR(iant1)->getChargeFactor();
          double eik = 0.0;
          double mu2 = 0.0;
          if ( iant1 <= iGGemitII && CMWtypeII > 0 ) {
            eik = (CMWtypeII == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121+s1j1+sj21)/s1j1/sj21);
            mu2 = aSkMu2I*s1j1*sj21/(S121+s1j1+sj21);
          } else if ( iant1 >= iQQemitIF && iant1 <= iGGemitIF && CMWtypeIF > 0 ) {
            eik = (CMWtypeIF == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121+sj21-s1j1)/s1j1/sj21);
            mu2 = aSkMu2I*s1j1*sj21/(S121+sj21);
          }
          double mu2def  = max(mu2min, mu2freeze + mu2);
          double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                           min(aSmax, aSptr->alphaS(mu2def));
          preFacCMW = eik*(coupFac-1.0);
        } else { // FSR
          antenna1  = getAntFSR(iant1)->antennaMassless(s1j1/S121,sj21/S121)/S121;
          antenna1 *= getAntFSR(iant1)->getChargeFactor();
          if ( iant1 <= iGGemit && CMWtypeFF > 0 ) {
            double eik     = (CMWtypeFF == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121-sj21-s1j1)/s1j1/sj21);
            double mu2     = aSkMu2F*s1j1*sj21/S121;
            double mu2def  = max(mu2min, mu2freeze + mu2);
            double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                             min(aSmax, aSptr->alphaS(mu2def));
            preFacCMW = eik*(coupFac-1.0);
          }
        }
      
        // Ordering with Pimp factor
        double Pimp1 = 1.0;
        double Q21   = scale21[iState1][iCl1];
        // Calculate Markovian ordering scale
        double Q2hat = 0.0;
        vector<int> iUnOrd;
        unorderingScale(clusteredParts1[iState1][iCl1], &Q2hat, &iUnOrd);
        Pimp1 = 1.0/(1.0+Q21/Q2hat);
        
        // Ariadne factor for final state gluon splitting
        double Pari1 = 1.0;
        if (doPari) if ( (iState1 == 0 && iant1 == iXGsplitIF) || 
                         (iState1 == 1 && iant1 == iQGsplit) || 
                         (iState1 == 1 && iant1 == iGGsplit) ) {
          int iGluon = 1; 
          int colG   = clusteredParts1[iState1][iCl1][iGluon].col();
          int acolG  = clusteredParts1[iState1][iCl1][iGluon].acol();
          int iN = -1;
          for (int i=2; i<(int)clusteredParts1[iState1][iCl1].size(); i++) { // Find colour partner
            // Start at 2 to leave out parents
            int colN  = clusteredParts1[iState1][iCl1][i].col();
            int acolN = clusteredParts1[iState1][iCl1][i].acol();
            if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
          }
          double sN = 0.0;
          if (iN >= 0) {
            sN = m2(clusteredParts1[iState1][iCl1][iGluon],clusteredParts1[iState1][iCl1][iN]);
            Pari1 = 2.0*sN/(S121 + sN);      
          }
        }
        
        // MEn
        double MEn1 = evalME(clusteredParts1[iState1][iCl1]);
        bool contrStrong1 = false;
        bool contrSmooth1 = (Q2hat > Q21);
        if (Q2hat > 0.0) { // should always be the case here
          // 0 strong ordering
          Q2clust1[0].push_back(Q21);
          if (Q2hat > Q21) {
            PSoverME[0] += ( Pari1 * antenna1 * MEn1 );
            nrOrdPaths[0]++;
            contrStrong1 = true;
          }
          // 1 smooth ordering
          Q2clust1[1].push_back(Q21);
          for (int i=0; i<(int)iUnOrd.size(); i++) {
            if (iUnOrd[i] == 0) contrSmooth1 = true;
            else if (iUnOrd[i] == 1) contrSmooth1 = true;
          }
          if (contrSmooth1) {
            PSoverME[1] += ( Pimp1 * Pari1 * antenna1 * MEn1 );
            nrOrdPaths[1]++;
          }
          // 2 no ordering
          Q2clust1[2].push_back(Q21);
          PSoverME[2] += ( Pari1 * antenna1 * MEn1 );
          nrOrdPaths[2]++;
        }
        
        // Add CMW contribution
        // TODO do we want to have the Pimp factor here?
        MEn2 += (preFacCMW*MEn1);
    
        if (print) {
          printOut("Vincia::MEcorrections::PSME3rdOrderFlat", 
                   "iAnt = " + num2str(iant1) +
                   " Q2 = " + num2str(Q21) +
                   " Q2hat = " + num2str(Q2hat) +
                   " Pimp = " + num2str(Pimp1) +
                   " Pari = " + num2str(Pari1) +
                   " (orderingMode=" + num2str(2) + ")");
          cout << "  Printing pre-branching momenta (" << (iState1 == 0 ? "ISR" : "FSR") 
               << "): id col acol p:" << endl;
          for (int i=0; i<(int)clusteredParts1[iState1][iCl1].size(); i++) {
            cout << "  " << setw(2) << clusteredParts1[iState1][iCl1][i].id();
            cout << " " << setw(3) << clusteredParts1[iState1][iCl1][i].col();
            cout << " " << setw(3) << clusteredParts1[iState1][iCl1][i].acol();
            cout << " " << clusteredParts1[iState1][iCl1][i].p();
          }
          cout << scientific;
          cout << "  antenna = " << antenna1 << " MEn+1 = " << MEn1;
          cout << " leadsto Pimp*Pari*antenna*MEn+1 = " << Pimp1*Pari1*antenna1*MEn1 << endl;
          if (preFacCMW > 0.0) 
            cout << "  CMW contribution to numerator = " << (preFacCMW*MEn1) << endl;
        }
        
        // Perform the second clustering
        clustering(clusteredParts1[iState1][iCl1]);
        // Copies
        vector< vector< vector<Particle> > > clusteredParts2 = clusteredPartsSav;
        vector< vector<        int       > > iAnts2          = iAntSav;
        vector< vector<        double    > > scale22         = scale2Sav;
        vector< vector< vector<double  > > > invariants2     = invariantsSav;
    
        // Loop over second clusterings
        for (int iState2=0; iState2<(int)clusteredParts2.size(); iState2++) if (iState2<2) {
          for (int iCl2=0; iCl2<(int)clusteredParts2[iState2].size(); iCl2++) {
            int iant2   = iAnts2[iState2][iCl2];
            double s1j2 = invariants2[iState2][iCl2][0];
            double sj22 = invariants2[iState2][iCl2][1];
            double S122 = invariants2[iState2][iCl2][3];
            double antenna2 = 0.0;
            if (iState2 == 0) { // ISR
              antenna2  = getAntISR(iant2)->antennaFunction(s1j2,sj22,S122);
              antenna2 *= getAntISR(iant2)->getChargeFactor();
            } else { // FSR
              antenna2  = getAntFSR(iant2)->antennaMassless(s1j2/S122,sj22/S122)/S122;
              antenna2 *= getAntFSR(iant2)->getChargeFactor();
            }
            
            // Ordering with Pimp factor
            double Pimp2 = 1.0;
            double Q22   = scale22[iState2][iCl2];
            unorderingScale(clusteredParts2[iState2][iCl2], &Q2hat, &iUnOrd);
            // Everyone is unordered anyway since only two emissions in second clustering
            Pimp2 = 1.0/(1.0+Q22/Q2hat);
            
            // Ariadne factor for final state gluon splitting
            double Pari2 = 1.0;
            if (doPari) if ( (iState2 == 0 && iant2 == iXGsplitIF) || 
                             (iState2 == 1 && iant2 == iQGsplit) || 
                             (iState2 == 1 && iant2 == iGGsplit) ) {
              int iGluon = 1; 
              int colG   = clusteredParts2[iState2][iCl2][iGluon].col();
              int acolG  = clusteredParts2[iState2][iCl2][iGluon].acol();
              int iN = -1;
              for (int i=2; i<(int)clusteredParts2[iState2][iCl2].size(); i++) { // Find colour partner
                // Start at 2 to leave out parents
                int colN  = clusteredParts2[iState2][iCl2][i].col();
                int acolN = clusteredParts2[iState2][iCl2][i].acol();
                if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
              }
              double sN = 0.0;
              if (iN >= 0) {
                sN = m2(clusteredParts2[iState2][iCl2][iGluon],clusteredParts2[iState2][iCl2][iN]);
                Pari2 = 2.0*sN/(S122 + sN);      
              }
            }
            
            // MEn
            double MEn = evalME(clusteredParts2[iState2][iCl2]);
            bool contrSmooth2 = (Q2hat > Q22);
            if (Q2hat > 0.0) { // should always be the case
              // 0 strong ordering
              Q2clust1[3].push_back(Q22);
              if (contrStrong1) if (Q2hat > Q22) {
                PSoverME[3] += ( Pari1 * Pari2 * antenna1 * antenna2 * MEn );
                nrOrdPaths[3]++;
              }
              // 1 smooth ordering
              Q2clust1[4].push_back(Q22);
              for (int i=0; i<(int)iUnOrd.size(); i++) {
                if (iUnOrd[i] == 0) contrSmooth2 = true;
                else if (iUnOrd[i] == 1) contrSmooth2 = true;
              }
              if (contrSmooth1 && contrSmooth2) {
                PSoverME[4] += ( Pimp1 * Pimp2 * Pari1 * Pari2 * antenna1 * antenna2 * MEn );
                nrOrdPaths[4]++;
              }
              // 2 no ordering
              Q2clust1[5].push_back(Q22);
              PSoverME[5] += ( Pari1 * Pari2 * antenna1 * antenna2 * MEn );
              nrOrdPaths[5]++;
            }
          }
        }
        // Loop over second clusterings
      }
    }
    // Loop over first clusterings
    
    for (int i=0; i<(int)PSoverME.size(); i++) 
      PSoverME[i] /= MEn2;
    
    if (print) printOut("Vincia::MEcorrections::PSME3rdOrderFlat", "MEn+2 = " + num2str(MEn2));
  }
  
  // Functions for flat phase space scan - fourth order
  
  void MEcorrections::PSME4thOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                                       vector< vector<double> >& Q2clust1,
                                       vector<int>& nrOrdPaths, bool print) {
    verbose = (print ? 9 : 1);
    PSoverME.clear();   PSoverME.resize(9);   // 0 strong ordering,
    Q2clust1.clear();   Q2clust1.resize(9);   // 1 smooth ordering,
    nrOrdPaths.clear(); nrOrdPaths.resize(9); // 2 no ordering
    for (int i=0; i<9; i++) {
      PSoverME[i] = 0.0;
      nrOrdPaths[i] = 0;
    }
    
    if (print) {
      printOut("Vincia::MEcorrections::PSME4thOrderFlat", "Printing post-branching momenta: id col acol p");
      for (int i=0; i<(int)parts.size(); i++) {
        cout << "  " << setw(2) << parts[i].id() << " " << setw(3) << parts[i].col();
        cout << " " << setw(3) << parts[i].acol() << " " << parts[i].p();
      }
    }
    
    // Mn+3
    double MEn3 = evalME(parts);
    // Perform the clustering
    clustering(parts);
    // Copies
    vector< vector< vector<Particle> > > clusteredParts1 = clusteredPartsSav;
    vector< vector<        int       > > iAnts1          = iAntSav;
    vector< vector<        double    > > scale21         = scale2Sav;
    vector< vector< vector<double  > > > invariants1     = invariantsSav;
    
    // Loop over first clusterings
    for (int iState1=0; iState1<(int)clusteredParts1.size(); iState1++) if (iState1<2) {
      for (int iCl1=0; iCl1<(int)clusteredParts1[iState1].size(); iCl1++) {
        int iant1   = iAnts1[iState1][iCl1];
        double s1j1 = invariants1[iState1][iCl1][0];
        double sj21 = invariants1[iState1][iCl1][1];
        double S121 = invariants1[iState1][iCl1][3];
        double antenna1  = 0.0;
        double preFacCMW = 0.0; // NOTE only sensible if matched to previous order
        if (iState1 == 0) { // ISR
          antenna1  = getAntISR(iant1)->antennaFunction(s1j1,sj21,S121);
          antenna1 *= getAntISR(iant1)->getChargeFactor();
          double eik = 0.0;
          double mu2 = 0.0;
          if ( iant1 <= iGGemitII && CMWtypeII > 0 ) {
            eik = (CMWtypeII == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121+s1j1+sj21)/s1j1/sj21);
            mu2 = aSkMu2I*s1j1*sj21/(S121+s1j1+sj21);
          } else if ( iant1 >= iQQemitIF && iant1 <= iGGemitIF && CMWtypeIF > 0 ) {
            eik = (CMWtypeIF == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121+sj21-s1j1)/s1j1/sj21);
            mu2 = aSkMu2I*s1j1*sj21/(S121+sj21);
          }
          double mu2def  = max(mu2min, mu2freeze + mu2);
          double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                           min(aSmax, aSptr->alphaS(mu2def));
          preFacCMW = eik*(coupFac-1.0);
        } else { // FSR
          antenna1  = getAntFSR(iant1)->antennaMassless(s1j1/S121,sj21/S121)/S121;
          antenna1 *= getAntFSR(iant1)->getChargeFactor();
          if ( iant1 <= iGGemit && CMWtypeFF > 0 ) {
            double eik     = (CMWtypeFF == 1 ? 2.0*S121/s1j1/sj21 : 2.0*(S121-sj21-s1j1)/s1j1/sj21);
            double mu2     = aSkMu2F*s1j1*sj21/S121;
            double mu2def  = max(mu2min, mu2freeze + mu2);
            double coupFac = min(aSmax, aSCMWptr->alphaS(mu2def)) /
                             min(aSmax, aSptr->alphaS(mu2def));
            preFacCMW = eik*(coupFac-1.0);
          }
        }
      
        // Ordering with Pimp factor
        double Pimp1 = 1.0;
        double Q21   = scale21[iState1][iCl1];
        // Calculate Markovian ordering scale
        double Q2hat = 0.0;
        vector<int> iUnOrd;
        unorderingScale(clusteredParts1[iState1][iCl1], &Q2hat, &iUnOrd);
        Pimp1 = 1.0/(1.0+Q21/Q2hat);
        
        // Ariadne factor for final state gluon splitting
        double Pari1 = 1.0;
        if (doPari) if ( (iState1 == 0 && iant1 == iXGsplitIF) || 
                         (iState1 == 1 && iant1 == iQGsplit) || 
                         (iState1 == 1 && iant1 == iGGsplit) ) {
          int iGluon = 1; 
          int colG   = clusteredParts1[iState1][iCl1][iGluon].col();
          int acolG  = clusteredParts1[iState1][iCl1][iGluon].acol();
          int iN = -1;
          for (int i=2; i<(int)clusteredParts1[iState1][iCl1].size(); i++) { // Find colour partner
            // Start at 2 to leave out parents
            int colN  = clusteredParts1[iState1][iCl1][i].col();
            int acolN = clusteredParts1[iState1][iCl1][i].acol();
            if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
          }
          double sN = 0.0;
          if (iN >= 0) {
            sN = m2(clusteredParts1[iState1][iCl1][iGluon],clusteredParts1[iState1][iCl1][iN]);
            Pari1 = 2.0*sN/(S121 + sN);      
          }
        }
        
        // MEn+2
        double MEn2 = evalME(clusteredParts1[iState1][iCl1]);
        bool contrStrong1 = false;
        bool contrSmooth1 = (Q2hat > Q21);
        if (Q2hat > 0.0) { // should always be the case here
          // 0 strong ordering
          Q2clust1[0].push_back(Q21);
          if (Q2hat > Q21) {
            PSoverME[0] += ( Pari1 * antenna1 * MEn2 );
            nrOrdPaths[0]++;
            contrStrong1 = true;
          }
          // 1 smooth ordering
          Q2clust1[1].push_back(Q21);
          for (int i=0; i<(int)iUnOrd.size(); i++) {
            if (iUnOrd[i] == 0) contrSmooth1 = true;
            else if (iUnOrd[i] == 1) contrSmooth1 = true;
          }
          if (contrSmooth1) {
            PSoverME[1] += ( Pimp1 * Pari1 * antenna1 * MEn2 );
            nrOrdPaths[1]++;
          }
          // 2 no ordering
          Q2clust1[2].push_back(Q21);
          PSoverME[2] += ( Pari1 * antenna1 * MEn2 );
          nrOrdPaths[2]++;
        }
        
        // Add CMW contribution
        // TODO do we want to have the Pimp factor here?
        MEn3 += (preFacCMW*MEn2);
    
        if (print) {
          printOut("Vincia::MEcorrections::PSME4thOrderFlat", 
                   "iAnt = " + num2str(iant1) +
                   " Q2 = " + num2str(Q21) +
                   " Q2hat = " + num2str(Q2hat) +
                   " Pimp = " + num2str(Pimp1) +
                   " Pari = " + num2str(Pari1) +
                   " (orderingMode=" + num2str(2) + ")");
          cout << "  Printing pre-branching momenta (" << (iState1 == 0 ? "ISR" : "FSR") 
               << "): id col acol p:" << endl;
          for (int i=0; i<(int)clusteredParts1[iState1][iCl1].size(); i++) {
            cout << "  " << setw(2) << clusteredParts1[iState1][iCl1][i].id();
            cout << " " << setw(3) << clusteredParts1[iState1][iCl1][i].col();
            cout << " " << setw(3) << clusteredParts1[iState1][iCl1][i].acol();
            cout << " " << clusteredParts1[iState1][iCl1][i].p();
          }
          cout << scientific;
          cout << "  antenna = " << antenna1 << " MEn+2 = " << MEn2;
          cout << " leadsto Pimp*Pari*antenna*MEn+2 = " << Pimp1*Pari1*antenna1*MEn2 << endl;
          if (preFacCMW > 0.0) 
            cout << "  CMW contribution to numerator = " << (preFacCMW*MEn2) << endl;
        }
        
        // Perform the second clustering
        clustering(clusteredParts1[iState1][iCl1]);
        // Copies
        vector< vector< vector<Particle> > > clusteredParts2 = clusteredPartsSav;
        vector< vector<        int       > > iAnts2          = iAntSav;
        vector< vector<        double    > > scale22         = scale2Sav;
        vector< vector< vector<double  > > > invariants2     = invariantsSav;
    
        // Loop over second clusterings
        for (int iState2=0; iState2<(int)clusteredParts2.size(); iState2++) if (iState2<2) {
          for (int iCl2=0; iCl2<(int)clusteredParts2[iState2].size(); iCl2++) {
            int iant2   = iAnts2[iState2][iCl2];
            double s1j2 = invariants2[iState2][iCl2][0];
            double sj22 = invariants2[iState2][iCl2][1];
            double S122 = invariants2[iState2][iCl2][3];
            double antenna2 = 0.0;
            if (iState2 == 0) { // ISR
              antenna2  = getAntISR(iant2)->antennaFunction(s1j2,sj22,S122);
              antenna2 *= getAntISR(iant2)->getChargeFactor();
            } else { // FSR
              antenna2  = getAntFSR(iant2)->antennaMassless(s1j2/S122,sj22/S122)/S122;
              antenna2 *= getAntFSR(iant2)->getChargeFactor();
            }
            
            // Ordering with Pimp factor
            double Pimp2 = 1.0;
            double Q22   = scale22[iState2][iCl2];
            unorderingScale(clusteredParts2[iState2][iCl2], &Q2hat, &iUnOrd);
            // Everyone is unordered anyway since only two emissions in second clustering
            Pimp2 = 1.0/(1.0+Q22/Q2hat);
            
            // Ariadne factor for final state gluon splitting
            double Pari2 = 1.0;
            if (doPari) if ( (iState2 == 0 && iant2 == iXGsplitIF) || 
                             (iState2 == 1 && iant2 == iQGsplit) || 
                             (iState2 == 1 && iant2 == iGGsplit) ) {
              int iGluon = 1; 
              int colG   = clusteredParts2[iState2][iCl2][iGluon].col();
              int acolG  = clusteredParts2[iState2][iCl2][iGluon].acol();
              int iN = -1;
              for (int i=2; i<(int)clusteredParts2[iState2][iCl2].size(); i++) { // Find colour partner
                // Start at 2 to leave out parents
                int colN  = clusteredParts2[iState2][iCl2][i].col();
                int acolN = clusteredParts2[iState2][iCl2][i].acol();
                if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
              }
              double sN = 0.0;
              if (iN >= 0) {
                sN = m2(clusteredParts2[iState2][iCl2][iGluon],clusteredParts2[iState2][iCl2][iN]);
                Pari2 = 2.0*sN/(S122 + sN);      
              }
            }
            
            // MEn+1
            double MEn1 = evalME(clusteredParts2[iState2][iCl2]);
            bool contrStrong2 = false;
            bool contrSmooth2 = (Q2hat > Q22);
            if (Q2hat > 0.0) { // should always be the case
              // 0 strong ordering
              Q2clust1[3].push_back(Q22);
              if (contrStrong1) if (Q2hat > Q22) {
                PSoverME[3] += ( Pari1 * Pari2 * antenna1 * antenna2 * MEn1 );
                nrOrdPaths[3]++;
                contrStrong2 = true;
              }
              // 1 smooth ordering
              Q2clust1[4].push_back(Q22);
              for (int i=0; i<(int)iUnOrd.size(); i++) {
                if (iUnOrd[i] == 0) contrSmooth2 = true;
                else if (iUnOrd[i] == 1) contrSmooth2 = true;
              }
              if (contrSmooth1 && contrSmooth2) {
                PSoverME[4] += ( Pimp1 * Pimp2 * Pari1 * Pari2 * antenna1 * antenna2 * MEn1 );
                nrOrdPaths[4]++;
              }
              // 2 no ordering
              Q2clust1[5].push_back(Q22);
              PSoverME[5] += ( Pari1 * Pari2 * antenna1 * antenna2 * MEn1 );
              nrOrdPaths[5]++;
            }
        
            // Perform the second clustering
            clustering(clusteredParts2[iState2][iCl2]);
            // Copies
            vector< vector< vector<Particle> > > clusteredParts3 = clusteredPartsSav;
            vector< vector<        int       > > iAnts3          = iAntSav;
            vector< vector<        double    > > scale23         = scale2Sav;
            vector< vector< vector<double  > > > invariants3     = invariantsSav;
    
            // Loop over third clusterings
            for (int iState3=0; iState3<(int)clusteredParts3.size(); iState3++) if (iState3<2) {
              for (int iCl3=0; iCl3<(int)clusteredParts3[iState3].size(); iCl3++) {
                int iant3   = iAnts2[iState3][iCl3];
                double s1j3 = invariants2[iState3][iCl3][0];
                double sj23 = invariants2[iState3][iCl3][1];
                double S123 = invariants2[iState3][iCl3][3];
                double antenna3 = 0.0;
                if (iState3 == 0) { // ISR
                  antenna3  = getAntISR(iant3)->antennaFunction(s1j3,sj23,S123);
                  antenna3 *= getAntISR(iant3)->getChargeFactor();
                } else { // FSR
                  antenna3  = getAntFSR(iant3)->antennaMassless(s1j3/S123,sj23/S123)/S123;
                  antenna3 *= getAntFSR(iant3)->getChargeFactor();
                }
            
                // Ordering with Pimp factor
                double Pimp3 = 1.0;
                double Q23   = scale23[iState3][iCl3];
                unorderingScale(clusteredParts3[iState3][iCl3], &Q2hat, &iUnOrd);
                // Everyone is unordered anyway since only two emissions in second clustering
                Pimp3 = 1.0/(1.0+Q23/Q2hat);
            
                // Ariadne factor for final state gluon splitting
                double Pari3 = 1.0;
                if (doPari) if ( (iState3 == 0 && iant3 == iXGsplitIF) || 
                                 (iState3 == 1 && iant3 == iQGsplit) || 
                                 (iState3 == 1 && iant3 == iGGsplit) ) {
                  int iGluon = 1; 
                  int colG   = clusteredParts3[iState3][iCl3][iGluon].col();
                  int acolG  = clusteredParts3[iState3][iCl3][iGluon].acol();
                  int iN = -1;
                  for (int i=2; i<(int)clusteredParts3[iState3][iCl3].size(); i++) { // Find colour partner
                    // Start at 2 to leave out parents
                    int colN  = clusteredParts3[iState3][iCl3][i].col();
                    int acolN = clusteredParts3[iState3][iCl3][i].acol();
                    if (colN == colG || acolN == acolG || colN == acolG || acolN == colG) iN = i;
                  }
                  double sN = 0.0;
                  if (iN >= 0) {
                    sN = m2(clusteredParts3[iState3][iCl3][iGluon],clusteredParts3[iState3][iCl3][iN]);
                    Pari3 = 2.0*sN/(S123 + sN);      
                  }
                }
            
                // MEn
                double MEn = evalME(clusteredParts3[iState3][iCl3]);
                bool contrSmooth3 = (Q2hat > Q23);
                if (Q2hat > 0.0) { // should always be the case
                  // 0 strong ordering
                  Q2clust1[6].push_back(Q23);
                  if (contrStrong1 && contrStrong2) if (Q2hat > Q23) {
                    PSoverME[6] += ( Pari1 * Pari2 * Pari3 * antenna1 * antenna2 * antenna3 * MEn );
                    nrOrdPaths[6]++;
                  }
                  // 1 smooth ordering
                  Q2clust1[7].push_back(Q23);
                  for (int i=0; i<(int)iUnOrd.size(); i++) {
                    if (iUnOrd[i] == 0) contrSmooth3 = true;
                    else if (iUnOrd[i] == 1) contrSmooth3 = true;
                  }
                  if (contrSmooth1 && contrSmooth2 && contrSmooth3) {
                    PSoverME[7] += ( Pimp1 * Pimp2 * Pimp3 * Pari1 * Pari2 * Pari3 * antenna1 * antenna2 * antenna3 * MEn );
                    nrOrdPaths[7]++;
                  }
                  // 2 no ordering
                  Q2clust1[8].push_back(Q23);
                  PSoverME[8] += ( Pari1 * Pari2 * Pari3 * antenna1 * antenna2 * antenna3 * MEn );
                  nrOrdPaths[8]++;
                }
              }
            }
            // Loop over third clusterings
          }
        }
        // Loop over second clusterings
      }
    }
    // Loop over first clusterings
    
    for (int i=0; i<(int)PSoverME.size(); i++) 
      PSoverME[i] /= MEn3;
    
    if (print) printOut("Vincia::MEcorrections::PSME4thOrderFlat", "MEn+3 = " + num2str(MEn3));
  }
  
  //************************************************************************
  
  // Perform the clustering
  
  void MEcorrections::clustering(vector<Particle> partonsIn) {
    clusteredPartsSav.clear(); clusteredPartsSav.resize(2);
    iAntSav.clear();           iAntSav.resize(2);
    scale2Sav.clear();         scale2Sav.resize(2);
    invariantsSav.clear();     invariantsSav.resize(2);
    for (int i=0; i<2; i++) {
      clusteredPartsSav[i].clear(); clusteredPartsSav[i].resize(0);
      iAntSav[i].clear();           iAntSav[i].resize(0);
      scale2Sav[i].clear();         scale2Sav[i].resize(0);
      invariantsSav[i].clear();     invariantsSav[i].resize(0);
    }
    
    // Get the colour chains (to check if gluon splitting/conversion is possible)
    int nQuarks = countNrQuarks(partonsIn);
    vector< vector<int> > colChains; 
    colChains.resize(0);
    if (nQuarks >= 4) colChains = getColChains(partonsIn);
    
    // Find and save all colors and anticolors
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    // Find initial guys
    int iA = -1;
    int iB = -1;
    for (int i=0; i<(int)partonsIn.size(); i++) {
      int col  = partonsIn[i].col();  // assuming that everything is
      int acol = partonsIn[i].acol(); // positiv here
      if (!partonsIn[i].isFinal()) { // Swap colours for initial guys
        col  = acol;
        acol = partonsIn[i].col();
        if ( (partonsIn[i].id()==21) || (partonsIn[i].idAbs()<7) ) {
          if (partonsIn[i].pz()>0.0) iA = i; else iB = i;
        }
      }
      if (col!=0)  indexOfCol[col]   = i;
      if (acol!=0) indexOfAcol[acol] = i;
    }
    
    // Only sensible if we found coloured initial state
    bool colIniState = (iA >= 0 && iB >= 0);
    Particle inA(partonsIn[max(iA,0)]);
    Particle inB(partonsIn[max(iB,0)]);
    int idA = 0; int colA = 0; int acolA = 0;
    int idB = 0; int colB = 0; int acolB = 0;
    bool ccAB = false;
    if (colIniState) {
      idA = inA.id(); colA = inA.col(); acolA = inA.acol();
      idB = inB.id(); colB = inB.col(); acolB = inB.acol();
      ccAB = ( ((colA==acolB)&&(colA!=0)) || ((acolA==colB)&&(acolA!=0)) );
    } else if (iA >= 0) {
      idA = inA.id(); colA = inA.col(); acolA = inA.acol();
      ccAB = false; 
    } else if (iB >= 0) {
      idB = inB.id(); colB = inB.col(); acolB = inB.acol();
      ccAB = false; 
    }
    
    // --------- Loop over all final state partons -- BEGIN
    for (int iJ=0; iJ<(int)partonsIn.size(); iJ++) if (partonsIn[iJ].isFinal()) {
      
      // The emission
      Particle outJ(partonsIn[iJ]);
      int idJ = outJ.id(); int colJ = outJ.col(); int acolJ = outJ.acol();
      
      // Container to store particles invariants later
      vector<double> invs; invs.resize(0);
      vector<Particle> out; out.resize(0);
      int iant  = 0;
      bool swap = false;
      
      // 1. Gluon emission ------------------------------------
      if (idJ == 21) {
        // Find the two colour partners
        int iP1 = indexOfCol[acolJ];
        int iP2 = indexOfAcol[colJ];
        // Swap positions, iP1 always initial and sideA if iP2 is initial as well
        if (!partonsIn[iP1].isFinal() && !partonsIn[iP2].isFinal()) {
          iP1 = iA; 
          iP2 = iB;
        } else if (!partonsIn[iP2].isFinal()) {
          int iPtmp = iP1;
          iP1 = iP2;
          iP2 = iPtmp;
        }
        Particle ioP1 = partonsIn[iP1];
        Particle ioP2 = partonsIn[iP2];
        int idP1 = partonsIn[iP1].id();
        int idP2 = partonsIn[iP2].id();
        
        // 1.1 from II
        if (!ioP1.isFinal() && !ioP2.isFinal()) {
          iant = iQQemitII;
          swap = (idP1!=21 && idP2==21);
               if (idP1==21 && idP2==21) iant = iGGemitII;
          else if (idP1==21 && idP2!=21) iant = iGQemitII;
          else if (idP1!=21 && idP2==21) iant = iGQemitII;
          out = clustII(ioP1, ioP2, outJ, iP1, iP2, iJ, partonsIn, iant, swap, &invs);
          if (out.size() == 0) continue;
          scale2Sav[0].push_back( resolutionPtr->Q2EII(ioP1, ioP2, outJ, evolTypeGemiI) );
          clusteredPartsSav[0].push_back( out );
          iAntSav[0].push_back( iant );
          invariantsSav[0].push_back( invs );
        }
        
        // 1.2 from IF
        else if (!ioP1.isFinal()) {
          iant = iQQemitIF;
          swap = false;
               if (idP1==21 && idP2==21) iant = iGGemitIF;
          else if (idP1==21 && idP2!=21) iant = iGQemitIF;
          else if (idP1!=21 && idP2==21) iant = iQGemitIF;
          out = clustIF(ioP1, ioP2, outJ, iP1, iP2, iJ, partonsIn, iant, &invs);
          if (out.size() == 0) continue;
          scale2Sav[0].push_back( resolutionPtr->Q2EIF(ioP1, ioP2, outJ, evolTypeGemiI) );
          clusteredPartsSav[0].push_back( out );
          iAntSav[0].push_back( iant );
          invariantsSav[0].push_back( invs );
        }
        
        // 1.3 from FF
        else {
          out = clustFF(ioP1, ioP2, outJ, iP1, iP2, iJ, partonsIn, &iant, &invs);
          if (out.size() == 0) continue;
          scale2Sav[1].push_back( resolutionPtr->Q2E(ioP1, outJ, ioP2, evolTypeGemiFF) );
          clusteredPartsSav[1].push_back( out );
          iAntSav[1].push_back( iant );
          invariantsSav[1].push_back( invs );
        }
      }
      
      // -- Non-Emission -- Start
      else if (abs(idJ) < 7) {
        // Find the colour partner
        int iP = (idJ > 0 ? indexOfAcol[colJ] : indexOfCol[acolJ]);
        Particle ioP = partonsIn[iP];
        int idP = partonsIn[iP].id();
        
        // 2. Initial gluon splitting ------------------------------------
        
        // 2.1 from II
        if (!ioP.isFinal() && idP==21 && ccAB) {
          iant = iQXsplitII;
          swap = (iP == iB);
          out = clustII(inA, inB, outJ, iA, iB, iJ, partonsIn, iant, swap, &invs);
          if (out.size() == 0) continue;
          scale2Sav[0].push_back( resolutionPtr->Q2EII(ioP, (iP==iA ? inB : inA), outJ, evolTypeGsplitI) );
          clusteredPartsSav[0].push_back( out );
          iAntSav[0].push_back( iant );
          invariantsSav[0].push_back( invs );
        }
        
        // 2.2 from IF
        else if (!ioP.isFinal() && idP==21) {
          int iK = indexOfAcol[ioP.acol()]; // Final state partner
          if (iK == iJ) iK = indexOfCol[ioP.col()];
          Particle outK(partonsIn[iK]);
          iant = iQXsplitIF;
          swap = false;
          out = clustIF(ioP, outK, outJ, iP, iK, iJ, partonsIn, iant, &invs);
          if (out.size() == 0) continue;
          scale2Sav[0].push_back( resolutionPtr->Q2EIF(ioP, outK, outJ, evolTypeGsplitI) );
          clusteredPartsSav[0].push_back( out );
          iAntSav[0].push_back( iant );
          invariantsSav[0].push_back( invs );
        }
        
        // Check if clustering of two quarks is ok
        if ((nQuarks-2) < 2*nQLmin) continue;
        
        // 3. Gluon conversion ------------------------------------
        // Find a initial quark with idJ and different colour
        vector<int> iPQIs; iPQIs.resize(0);
        if ((idA == idJ) && (iA != iP) && canClustQQB(partonsIn, colChains, iJ, iA))
          iPQIs.push_back(iA);
        if ((idB == idJ) && (iB != iP) && canClustQQB(partonsIn, colChains, iJ, iB))
          iPQIs.push_back(iB);
        
        // Loop over all possible partner quarks
        for (int ipqi=0; ipqi<(int)iPQIs.size(); ipqi++) {
          // The partner quark
          int iPQI = iPQIs[ipqi];
          Particle inPQ(partonsIn[iPQI]);
          
          // 3.1 from II with J as emission
          if (!ioP.isFinal()) {
            iant = iGXconvII;
            swap = (iPQI == iB);
            out = clustII(inA, inB, outJ, iA, iB, iJ, partonsIn, iant, swap, &invs);
            if (out.size() == 0) continue;
            scale2Sav[0].push_back( resolutionPtr->Q2EII(inPQ, ioP, outJ, evolTypeQconvI) );
            clusteredPartsSav[0].push_back( out );
            iAntSav[0].push_back( iant );
            invariantsSav[0].push_back( invs );
          }
          
          // 3.2 from IF with J as emission
          else {
            iant = iGXconvIF;
            swap = false;
            out = clustIF(inPQ, ioP, outJ, iPQI, iP, iJ, partonsIn, iant, &invs);
            if (out.size() == 0) continue;
            scale2Sav[0].push_back( resolutionPtr->Q2EIF(inPQ, ioP, outJ, evolTypeQconvI) );
            clusteredPartsSav[0].push_back( out );
            iAntSav[0].push_back( iant );
            invariantsSav[0].push_back( invs );
          }
          
          // 3.3 from II with iPQI as emission
          if (ccAB) {
            iant = iGXconvII;
            swap = (iPQI == iB);
            out = clustII(inA, inB, outJ, iA, iB, iJ, partonsIn, iant, swap, &invs);
            if (out.size() == 0) continue;
            scale2Sav[0].push_back( resolutionPtr->Q2EII(inPQ, ioP, outJ, evolTypeQconvI) );
            clusteredPartsSav[0].push_back( out );
            iAntSav[0].push_back( iant );
            invariantsSav[0].push_back( invs );
          }
          
          // 3.4 from IF with iPQI as emission
          else {
            int iK = (inPQ.id() > 0 ? indexOfCol[inPQ.col()]
                                    : indexOfAcol[inPQ.acol()]); // Final state partner of initial quark
            Particle outK(partonsIn[iK]);
            iant = iGXconvIF;
            swap = false;
            out = clustIF(inPQ, outK, outJ, iPQI, iK, iJ, partonsIn, iant, &invs);
            if (out.size() == 0) continue;
            scale2Sav[0].push_back( resolutionPtr->Q2EIF(inPQ, outK, outJ, evolTypeQconvI) );
            clusteredPartsSav[0].push_back( out );
            iAntSav[0].push_back( iant );
            invariantsSav[0].push_back( invs );
          }
        }
        
        // 4. Final gluon splitting ------------------------------------
        // Find a final quark with -idJ and different colour
        vector<int> iPQs; iPQs.resize(0);
        for (int iPQ=0; iPQ<(int)partonsIn.size(); iPQ++) {
          Particle outPQ(partonsIn[iPQ]);
          if (outPQ.isFinal() && outPQ.id() == -idJ && iPQ != iP) {
            if (canClustQQB(partonsIn, colChains, iJ, iPQ))
              iPQs.push_back(iPQ);
          }
        }
        // Loop over all possible partner quarks
        for (int ipq=0; ipq<(int)iPQs.size(); ipq++) {
          // The partner quark
          int iPQ = iPQs[ipq];
          Particle outPQ(partonsIn[iPQ]);
          
          // 4.1 from IF
          if (!ioP.isFinal()) {
            iant = iXGsplitIF;
            swap = false;
            out = clustIF(ioP, outPQ, outJ, iP, iPQ, iJ, partonsIn, iant, &invs);
            if (out.size() == 0) continue;
            scale2Sav[0].push_back( resolutionPtr->Q2EIF(ioP, outPQ, outJ, evolTypeGsplitF) );
            clusteredPartsSav[0].push_back( out );
            iAntSav[0].push_back( iant );
            invariantsSav[0].push_back( invs );
          }
          
          // 4.2 from FF
          else {
            if (idJ > 0) out = clustFF(outPQ, ioP, outJ, iPQ, iP, iJ, partonsIn, &iant, &invs);
            else         out = clustFF(ioP, outPQ, outJ, iP, iPQ, iJ, partonsIn, &iant, &invs);
            if (out.size() == 0) continue;
            scale2Sav[1].push_back( resolutionPtr->Q2E(ioP, outJ, outPQ, evolTypeGsplitFF) );
            clusteredPartsSav[1].push_back( out );
            iAntSav[1].push_back( iant );
            invariantsSav[1].push_back( invs );
          }
        }
      }
      // -- Non-Emission -- END
      
    }
    // --------- Loop over all final state partons -- END
  }
  
  //************************************************************************
  
  // Get the clustered momenta
  
  vector<Particle> MEcorrections::clustII(Particle ina, Particle inb, Particle inj, 
                                          int ia, int ib, int ij,
                                          vector<Particle> partonsIn, int iAnt, bool swapped, 
                                          vector<double>* invs) {
    // Reset
    vector<Particle> out; 
    out.resize(0);
    invs->clear(); 
    invs->resize(0);
    
    // Invariants
    double sab = (ina.p()+inb.p()).m2Calc();
    double saj = (ina.p()+inj.p()).m2Calc();
    double sjb = (inj.p()+inb.p()).m2Calc();
    double sAB = sab - saj - sjb;
    // We can't to the clustering if sAB < 0
    if (sAB < 0.0) return out;
    if (!swapped) { 
      invs->push_back(saj);
      invs->push_back(sjb);
    } else { 
      invs->push_back(sjb);
      invs->push_back(saj);
    }
    invs->push_back(sab);
    invs->push_back(sAB);
    
    // Clustering: kinematic part
    Vec4 pA = ina.p();
    Vec4 pB = inb.p();
    double rescaleFacA = sqrt(sAB/sab*(sAB+saj)/(sAB+sjb));
    double rescaleFacB = sqrt(sAB/sab*(sAB+sjb)/(sAB+saj));
    pA.rescale4(rescaleFacA);
    pB.rescale4(rescaleFacB);
    // Check energys -> is clustering possible?
    if ( (pA.e() < 0.0) || (pA.e() > ina.e()) ) return out;
    if ( (pB.e() < 0.0) || (pB.e() > inb.e()) ) return out;
    
    // Id's and colours
    int idA   = ina.id();
    int colA  = ina.col();
    int acolA = ina.acol();
    int idB   = inb.id();
    int colB  = inb.col();
    int acolB = inb.acol();
    int idj   = inj.id();
    int colj  = inj.col();
    int acolj = inj.acol();
    if (idj == 21) {                                // 1) Gluon emission
      if (colj == colB) colB = acolA;
      else if (acolj == acolB) acolB = colA;
    } else if ( (iAnt == iGXconvII) && !swapped) {  // 2) Initital state gluon conversion sideA
      idA = 21;
      if (colA == 0) colA = acolj;
      if (acolA == 0) acolA = colj;
    } else if ( (iAnt == iGXconvII) &&  swapped) {  // 3) Initital state gluon conversion sideB
      idB = 21;
      if (colB == 0) colB = acolj;
      if (acolB == 0) acolB = colj;
    } else if ( (iAnt == iQXsplitII) && !swapped) { // 4) Initital state gluon splitting side A
      idA = -idj;
      if (idA > 0) acolA = 0;
      else if (idA < 0) colA = 0;
    } else if ( (iAnt == iQXsplitII) &&  swapped) { // 5) Initital state gluon splitting side B
      idB = -idj;
      if (idB > 0) acolB = 0;
      else if (idB < 0) colB = 0;
    }
    
    // Particle vector for recoiler
    vector<Particle> recoiler; 
    recoiler.resize(0);
    for (int i=0; i<(int)partonsIn.size(); i++) {
      if ((i!=ij) && (i!=ia) && (i!=ib)) recoiler.push_back(partonsIn[i]);
    }
    // 1) Boost in center-of-mass frame AB, from E,0,0,pz to E,0,0,0
    Vec4 pSum = pA + pB;
    // 2) Boost from E,0,0,0 to E',px',py',pz' (recoiler before clustering in lab frame)
    Vec4 pRec(0.0, 0.0, 0.0, 0.0);
    for (int i=0; i<(int)recoiler.size(); i++) pRec += recoiler[i].p();
    // Now we perform the boost back
    for (int i=0; i<(int)recoiler.size(); i++) {
      recoiler[i].bstback(pRec);
      recoiler[i].bst(pSum);
    }
    
    // Fill with new guys
    if (!swapped) { 
      out.push_back( Particle(idA, -21, 0, 0, 0, 0, colA, acolA, pA) );
      out.push_back( Particle(idB, -21, 0, 0, 0, 0, colB, acolB, pB) );
    } else {
      out.push_back( Particle(idB, -21, 0, 0, 0, 0, colB, acolB, pB) );
      out.push_back( Particle(idA, -21, 0, 0, 0, 0, colA, acolA, pA) );
    }
    for (int i=0; i<(int)recoiler.size(); i++) out.push_back( recoiler[i] );
    
    // Done!
    return out;
  }
  
  vector<Particle> MEcorrections::clustIF(Particle ina, Particle ink, Particle inj, 
                                          int ia, int ik, int ij,
                                          vector<Particle> partonsIn, int iAnt, 
                                          vector<double>* invs) {
    // Reset
    vector<Particle> out; 
    out.resize(0);
    invs->clear(); 
    invs->resize(0);
    
    // Invariants
    double sak = (ina.p()+ink.p()).m2Calc();
    double saj = (ina.p()+inj.p()).m2Calc();
    double sjk = (inj.p()+ink.p()).m2Calc();
    double sAK = sak + saj - sjk;
    // We can't to the clustering if sAK < 0
    if (sAK < 0.0) return out;
    invs->push_back(saj); 
    invs->push_back(sjk); 
    invs->push_back(sak); 
    invs->push_back(sAK);
    
    // Clustering: kinematic part (pA - pK = pa - pk -pj)
    Vec4 pA = ina.p();
    pA.rescale4(sAK/(sAK+sjk));
    Vec4 pK = pA-ina.p()+ink.p()+inj.p();
    // Check energys -> is clustering possible?
    if ( (pA.e() < 0.0) || (pK.e() < 0.0) ) return out;
    
    // Id's and colours
    int idA    = ina.id();
    int colA   = ina.col();
    int acolA  = ina.acol();
    int idK    = ink.id();
    int colK   = ink.col();
    int acolK  = ink.acol();
    int idj    = inj.id();
    int colj   = inj.col();
    int acolj  = inj.acol();
    if (idj == 21) {                 // 1) Gluon emission
      if (colj == acolK) acolK = acolA;
      else if (acolj == colK) colK = colA;
    } else if (iAnt == iXGsplitIF) { // 2) Final state gluon splitting
      idK = 21;
      if (colK == 0) colK = colj;
      if (acolK == 0) acolK = acolj;
    } else if (iAnt == iGXconvIF) {  // 3) Initital state gluon conversion
      idA = 21;
      if (colA == 0) colA = acolj;
      if (acolA == 0) acolA = colj;
    } else if (iAnt == iQXsplitIF) { // 4) Initital state gluon splitting
      idA = -idj;
      if (idA > 0) acolA = 0;
      else if (idA < 0) colA = 0;
    }
    
    // Fill with new guys
    out.push_back( Particle(idA, -21, 0, 0, 0, 0, colA, acolA, pA) );
    out.push_back( Particle(idK,  51, 0, 0, 0, 0, colK, acolK, pK) );
    for (int i=0; i<(int)partonsIn.size(); i++) 
      if (i!=ia && i!=ik && i!=ij) out.push_back( partonsIn[i] );
    
    // Done!
    return out;
  }
    
  vector<Particle> MEcorrections::clustFF(Particle ini, Particle ink, Particle inj,
                                          int ii, int ik, int ij,
                                          vector<Particle> partonsIn, int* iAnt,
                                          vector<double>* invs) {
    // Reset
    vector<Particle> out; 
    out.resize(0);
    invs->clear(); 
    invs->resize(0);
    
    // Invariants
    double sik = (ini.p()+ink.p()).m2Calc();
    double sij = (ini.p()+inj.p()).m2Calc();
    double sjk = (inj.p()+ink.p()).m2Calc();
    double sIK = sik + sij + sjk;
    
    // Clustering: kinematic part
    vector<int> idsVc;
    idsVc.push_back(ini.id());
    idsVc.push_back(inj.id());
    idsVc.push_back(ink.id());
    vector<Vec4> pVc;
    pVc.push_back(ini.p());
    pVc.push_back(inj.p());
    pVc.push_back(ink.p());
    int ia = 0;
    int ir = 1;
    int ib = 2;
    vector<int> idClu; 
    vector<Vec4> pClu;
    int iant;
    bool swapAB;
    if (!vinCluPtr->getClu(idsVc, pVc, ia, ir, ib, idClu,
                           pClu, iant, swapAB)) return out;
    
    // Id's and colours
    int idj   = inj.id();
    int colj  = inj.col();
    int acolj = inj.acol();
    int idI   = idClu[0];
    int colI  = ini.col();
    int acolI = ini.acol();
    int idK   = idClu[1];
    int colK  = ink.col();
    int acolK = ink.acol();
    // 1) Gluon emission
    if (idj == 21) {
      if (colj == acolK) acolK = colI;
      else if (acolj == colK) colK = acolI;
    }
    // 2) Gluon splitting on side I (default is g = jk)
    else if (swapAB) {
      idI = 21;
      if (colI == 0) colI = colj;
      if (acolI == 0) acolI = acolj;
    }
    // 3) Gluon splitting on side K (default is g = jk)
    else {
      idK = 21;
      if (colK == 0) colK = colj;
      if (acolK == 0) acolK = acolj;
    }
    
    *iAnt = iant;
    if (!swapAB) {
      invs->push_back(sij);
      invs->push_back(sjk);
    } else {
      invs->push_back(sjk);
      invs->push_back(sij);
    }
    invs->push_back(sik);
    invs->push_back(sIK);
    
    // Fill with new guys
    if (swapAB) { // Put g that splitted in second to make Pari easier
      out.push_back( Particle(idK, 51, 0, 0, 0, 0, colK, acolK, pClu[1]) );
      out.push_back( Particle(idI, 51, 0, 0, 0, 0, colI, acolI, pClu[0]) );
    }
    else {
      out.push_back( Particle(idI, 51, 0, 0, 0, 0, colI, acolI, pClu[0]) );
      out.push_back( Particle(idK, 51, 0, 0, 0, 0, colK, acolK, pClu[1]) );
    }
    for (int i=0; i<(int)partonsIn.size(); i++) 
      if (i!=ii && i!=ik && i!=ij) out.push_back( partonsIn[i] );
      
    // Done!
    return out;
  }
  
  //************************************************************************
  
  // Get the order of the n partons for evaluating the matrix elements
  
  vector< vector<int> > MEcorrections::orderForME(vector<Particle> parts) {
    vector<int> order; order.clear(); order.resize(0);
    vector<int> posII; posII.clear(); posII.resize(0);
    
    int nQuarks  = countNrQuarks(parts);
    int nGluons  = countNrGluons(parts);
    
    if (nQuarks == 0) { //  Special case: gluons only, start with gi1
      
      // 2 gluons only
      if (nGluons == 2) {
        for (int i=0; i<(int)parts.size(); i++)
          if (parts[i].id() == 21) order.push_back(i);
        posII.push_back(0);
        posII.push_back(1);
        vector< vector<int> > combi; combi.clear(); combi.resize(0);
        combi.push_back(order); combi.push_back(posII);
        return combi;
      }
      
      int iG1 = -1;
      // Find and save all colours and anticolours
      map<int,int> indexOfAcol; // Put initial colours here
      map<int,int> indexOfCol;  // Put initial anticolours here
      for (int i=0; i<(int)parts.size(); i++) {
        int col   = 0;
        int acol  = 0;
        bool isIS = !parts[i].isFinal();
        if (isIS) {
          col  = parts[i].acol();
          acol = parts[i].col();
        }
        else {
          col  = parts[i].col();
          acol = parts[i].acol();
        }
        if (col > 0) indexOfCol[col] = i;
        else if (col < 0) indexOfAcol[-col] = i;
        if (acol > 0) indexOfAcol[acol] = i;
        else if (acol < 0) indexOfCol[-acol] = i;
        if ( (parts[i].id() == 21) && (isIS) && (iG1 == -1) ) iG1 = i;
      }
      
      int col = 0;
      // 1) Initial state gluons present
      if (iG1 > -1) {
        col = parts[iG1].acol();
        posII.push_back(0);
      }
      // 2) No initial state gluons
      else {
        // Pick random gluon
        for (int i=0; i<(int)parts.size(); i++)  if (parts[i].id() == 21) {
          iG1 = i;
          col = parts[iG1].col();
          break;
        }
      }      
      
      order.push_back(iG1);
      bool done  = false;
      int iGnext = -1;
      int count  = 0;
      while (!done) {
        count++;
        iGnext = indexOfAcol[col];
        if (iGnext == iG1) done = true; // Found the end (gluon we started with)
        else {
          order.push_back(iGnext);
          if (!parts[iGnext].isFinal()) {
            posII.push_back(count);
            col = parts[iGnext].acol(); // Update colour to look for
          } else col = parts[iGnext].col();
        }
      }
      
      vector< vector<int> > combi; combi.clear(); combi.resize(0);
      combi.push_back(order); combi.push_back(posII);
      return combi;
        
    }
    
    vector< vector<int> > cChains;
    vector<int> iQs;
    // Find and save all colours and anticolours
    map<int,int> indexOfAcol; // Put initial colours here
    map<int,int> indexOfCol;  // Put initial anticolours here
    for (int i=0; i<(int)parts.size(); i++) {
      int col  = 0;
      int acol = 0;
      bool isIS = !parts[i].isFinal();
      if (isIS) {
        col  = parts[i].acol();
        acol = parts[i].col();
      }
      else {
        col  = parts[i].col();
        acol = parts[i].acol();
      }
      if (col > 0) indexOfCol[col] = i;
      else if (col < 0) indexOfAcol[-col] = i;
      if (acol > 0) indexOfAcol[acol] = i;
      else if (acol < 0) indexOfCol[-acol] = i;
      int id = parts[i].id();
      if (abs(id) < 6) {
        if ( (id > 0) && (!isIS) ) iQs.push_back(i);
        else if ( (id < 0) && (isIS) ) iQs.push_back(i);
      }
    }
    
    // Loop over all quarks
    for (int iQ=0; iQ<(int)iQs.size(); iQ++) {
      vector<int> chain; chain.clear(); chain.resize(0);
      chain.push_back(iQs[iQ]);
      int col    = (parts[iQs[iQ]].isFinal() ? parts[iQs[iQ]].col() : parts[iQs[iQ]].acol());
      bool done  = false;
      int iPnext = -1;
      while (!done) {
        iPnext = indexOfAcol[col];
        chain.push_back(iPnext);
        int id = parts[iPnext].id();
        if (parts[iPnext].isFinal()) {
          if ( (id < 0) &&  (abs(id) < 6) ) done = true; // Found the end of the chain (antiquark)
          col = parts[iPnext].col(); // Update colour to look for
        }
        else {
          if ( (id > 0) &&  (abs(id) < 6) ) done = true; // Found the end of the chain (antiquark)
          col = parts[iPnext].acol(); // Update colour to look for
        }
      }
      cChains.push_back(chain);
    }
    
    // Sort vectors in size
    sort(cChains.begin(), cChains.end(), compVecSizeHelper);
    
    // Loop over the chains and put nr in result
    int count = 0;
    for (int i=0; i<(int)cChains.size(); i++) {
      for (int j=0; j<(int)cChains[i].size(); j++) {
        order.push_back(cChains[i][j]);
        if (!parts[cChains[i][j]].isFinal()) posII.push_back(count);
        count++;
      }
    }
    
    vector< vector<int> > combi; combi.clear(); combi.resize(0);
    combi.push_back(order); combi.push_back(posII);
    return combi;
  }
  
  //************************************************************************
  
  // Evaluate a matrix element for the defined process classes
  
  double MEcorrections::evalME(vector<Particle> parts) {
    // Safety check
    if (!canDoMEC(parts, false)) {
      if (verbose >= 4) printOut("Vincia::MEcorrections::evalME",
                                 "Requested ME that we don't have! Returning 0!");
      return 0.0;
    }
    
    vector<int> MGids; MGids.clear(); MGids.resize(0); // IDs
    vector<Vec4> p; p.clear(); p.resize(0); // Input Momenta
    vector< vector<int> > order = orderForME(parts);
    if (order[0].size()==0) return -1.0;
    
    // Fill IDs and momenta, check if we have single boson
    int nIn      = ( decayME() ? 1 : 2 );
    int iBoson   = -1;
    if (decayME() || prodME()) {
      for (int i=0; i<(int)parts.size(); i++) 
        if ( (parts[i].idAbs() > 21) && (parts[i].idAbs() < 38) ) iBoson = i;
    }
    if ( decayME() && (iBoson != -1) ) {
      MGids.push_back(parts[iBoson].id());
      p.push_back(parts[iBoson].p());
    }
    for (int j=0; j<(int)order[0].size(); j++) {
      MGids.push_back(parts[order[0][j]].id());
      p.push_back(parts[order[0][j]].p());
    }
    if ( prodME() && (iBoson != -1) ) {
      MGids.push_back(parts[iBoson].id());
      p.push_back(parts[iBoson].p());
    }
    
    // Position of incoming guys
    if (twoToN()) mgInterfacePtr->setIIpos( order[1][0], order[1][1] );
    
    if (verbose >= 5) {
      printOut("Vincia::MEcorrections::evalME","Printing ids and momenta");
      for (int i=0; i<(int)MGids.size(); i++) cout << "  " << setw(2) << MGids[i] << " " << p[i];
      if (twoToN()) cout << "  ipos1 = " << order[1][0] << " and ipos2 = " << order[1][1] << endl;
    }
    
    return mgInterfacePtr->ME2(nIn,MGids,p);
  }
  
  //************************************************************************
  
  // Function for the unordering: Finds the smallest scale and a list
  // of partons that should be unordered
  
  void MEcorrections::unorderingScale(vector<Particle> partonsIn, double* Q2min, 
                                      vector<int>* unOrdPartons) {
    
    // Reset
    *Q2min = 0.0;
    double Q2large = pow(10.0,100.0);
    unOrdPartons->resize(0);
    vector<int> iUnOrd; iUnOrd.resize(0);
    
    // Get the colour chains (to check if gluon splitting/conversion is possible)
    int nQuarks = countNrQuarks(partonsIn);
    vector< vector<int> > colChains; 
    colChains.resize(0);
    if (nQuarks >= 4) colChains = getColChains(partonsIn);
    
    // Find and save all colors and anticolors
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    // Find initial guys
    int iA = -1;
    int iB = -1;
    for (int i=0; i<(int)partonsIn.size(); i++) {
      int col  = partonsIn[i].col();  // assuming that everything is
      int acol = partonsIn[i].acol(); // positiv here
      if (!partonsIn[i].isFinal()) { // Swap colours for initial guys
        col  = acol;
        acol = partonsIn[i].col();
        if ( (partonsIn[i].id()==21) || (abs(partonsIn[i].id())<7) ) {
          if (partonsIn[i].pz()>0.0) iA = i; else iB = i;
        }
      }
      if (col!=0)  indexOfCol[col]   = i;
      if (acol!=0) indexOfAcol[acol] = i;
    }
    
    // Only sensible if we found coloured initial state
    bool colIniState = (iA >= 0 && iB >= 0);
    Particle inA(partonsIn[max(iA,0)]);
    Particle inB(partonsIn[max(iB,0)]);
    int idA = 0; int colA = 0; int acolA = 0;
    int idB = 0; int colB = 0; int acolB = 0;
    bool ccAB = false;
    if (colIniState) {
      idA = inA.id(); colA = inA.col(); acolA = inA.acol();
      idB = inB.id(); colB = inB.col(); acolB = inB.acol();
      ccAB = ( ((colA==acolB)&&(colA!=0)) || ((acolA==colB)&&(acolA!=0)) );
    } else if (iA >= 0) {
      idA = inA.id(); colA = inA.col(); acolA = inA.acol();
      ccAB = false; 
    } else if (iB >= 0) {
      idB = inB.id(); colB = inB.col(); acolB = inB.acol();
      ccAB = false; 
    }
    
    // --------- Loop over all final state partons -- BEGIN
    for (int iJ=0; iJ<(int)partonsIn.size(); iJ++) if (partonsIn[iJ].isFinal()) {
      
      // The emission
      Particle outJ(partonsIn[iJ]);
      int idJ = outJ.id(); int colJ = outJ.col(); int acolJ = outJ.acol();
      
      // 1. Gluon emission ------------------------------------
      if (idJ == 21) {
        // Find the two colour partners
        int iP1 = indexOfCol[acolJ];
        int iP2 = indexOfAcol[colJ];
        // Swap positions, iP1 always initial and sideA if iP2 is initial as well
        if (!partonsIn[iP1].isFinal() && !partonsIn[iP2].isFinal()) {
          iP1 = iA; 
          iP2 = iB;
        } else if (!partonsIn[iP2].isFinal()) {
          int iPtmp = iP1;
          iP1 = iP2;
          iP2 = iPtmp;
        }
        Particle ioP1 = partonsIn[iP1];
        Particle ioP2 = partonsIn[iP2];
        double QE2 = 0.0;
        // 1.1 from II
        if (!ioP1.isFinal() && !ioP2.isFinal())
          QE2 = resolutionPtr->Q2EII(ioP1, ioP2, outJ, evolTypeGemiI);
        // 1.2 from IF
        else if (!ioP1.isFinal())
          QE2 = resolutionPtr->Q2EIF(ioP1, ioP2, outJ, evolTypeGemiI);
        // 1.3 from FF
        else
          QE2 = resolutionPtr->Q2E(ioP1, outJ, ioP2, evolTypeGemiFF);
        if (QE2 < Q2large) { iUnOrd.resize(0);     
                             iUnOrd.push_back(iJ); 
                             iUnOrd.push_back(iP1); 
                             iUnOrd.push_back(iP2);
                             Q2large = QE2; }
      }
      
      // -- Non-Emission -- Start
      else if (abs(idJ) < 7) {
        // Find the colour partner
        int iP = (idJ > 0 ? indexOfAcol[colJ] : indexOfCol[acolJ]);
        Particle ioP = partonsIn[iP];
        int idP = partonsIn[iP].id();
        
        // 2. Initial gluon splitting ------------------------------------
        // 2.1 from II
        if (!ioP.isFinal() && idP==21 && ccAB) {
          double QE2 = resolutionPtr->Q2EII(ioP, (iP==iA ? inB : inA), outJ, evolTypeGsplitI);
          if (QE2 < Q2large) { iUnOrd.resize(0);     
                               iUnOrd.push_back(iJ); 
                               iUnOrd.push_back(iA); 
                               iUnOrd.push_back(iB);
                               Q2large = QE2; }
        }
        // 2.2 from IF
        else if (!ioP.isFinal() && idP==21) {
          int iK = indexOfAcol[ioP.acol()]; // Final state partner
          if (iK == iJ) iK = indexOfCol[ioP.col()];
          Particle outK(partonsIn[iK]);
          double QE2 = resolutionPtr->Q2EIF(ioP, outK, outJ, evolTypeGsplitI);
          if (QE2 < Q2large) { iUnOrd.resize(0);     
                               iUnOrd.push_back(iJ); 
                               iUnOrd.push_back(iP); 
                               iUnOrd.push_back(iK);
                               Q2large = QE2; }
        }
        
        // Check if clustering of two quarks is ok
        if ((nQuarks-2) < 2*nQLmin) continue;
        
        // 3. Gluon conversion ------------------------------------
        vector<int> iPQIs; iPQIs.resize(0);
        if ((idA == idJ) && (iA != iP) && canClustQQB(partonsIn, colChains, iJ, iA))
          iPQIs.push_back(iA);
        if ((idB == idJ) && (iB != iP) && canClustQQB(partonsIn, colChains, iJ, iB))
          iPQIs.push_back(iB);
        // Loop over all possible partner quarks
        for (int ipqi=0; ipqi<(int)iPQIs.size(); ipqi++) {
          // The partner quark
          int iPQI = iPQIs[ipqi];
          Particle inPQ(partonsIn[iPQI]);
          // Colour partner of initial partner quark
          int iCPPQ = (inPQ.id() > 0 ? indexOfCol[inPQ.col()] : indexOfAcol[inPQ.acol()]);
          Particle ioCPPQ(partonsIn[iCPPQ]);
          double QE2 = 0.0;
          // 3.1 from II with J as emission
          if (!ioP.isFinal()) 
            QE2 = resolutionPtr->Q2EII(inPQ, ioP, outJ, evolTypeQconvI);
          // 3.2 from IF with J as emission
          else
            QE2 = resolutionPtr->Q2EIF(inPQ, ioP, outJ, evolTypeQconvI);
          if (QE2 < Q2large) { iUnOrd.resize(0);
                               iUnOrd.push_back(iJ);
                               iUnOrd.push_back(iP);       
                               iUnOrd.push_back(iPQI); 
                               iUnOrd.push_back(iCPPQ); 
                               Q2large = QE2; }
          // 3.3 from II with iPQI as emission
          if (ccAB)
            QE2 = resolutionPtr->Q2EII(inPQ, ioP, outJ, evolTypeQconvI);
          // 3.4 from IF with iPQI as emission
          else
            QE2 = resolutionPtr->Q2EIF(inPQ, ioCPPQ, outJ, evolTypeQconvI);
          if (QE2 < Q2large) { iUnOrd.resize(0);
                               iUnOrd.push_back(iJ);
                               iUnOrd.push_back(iP);       
                               iUnOrd.push_back(iPQI); 
                               iUnOrd.push_back(iCPPQ);
                               Q2large = QE2; }
        }
        
        // 4. Final gluon splitting ------------------------------------
        // Find a final quark with -idJ and different colour
        vector<int> iPQs; iPQs.resize(0);
        for (int iPQ=0; iPQ<(int)partonsIn.size(); iPQ++) {
          Particle outPQ(partonsIn[iPQ]);
          if (outPQ.isFinal() && outPQ.id() == -idJ && iPQ != iP) {
            if (canClustQQB(partonsIn, colChains, iJ, iPQ))
              iPQs.push_back(iPQ);
          }
        }
        // Loop over all possible partner quarks
        for (int ipq=0; ipq<(int)iPQs.size(); ipq++) {
          // The partner quark
          int iPQ = iPQs[ipq];
          Particle outPQ(partonsIn[iPQ]);
          // Colour partner of final partner quark
          int iCPPQ = (outPQ.id() > 0 ? indexOfAcol[outPQ.col()] : indexOfCol[outPQ.acol()]);
          Particle ioCPPQ(partonsIn[iCPPQ]);
          double QE2 = 0.0;
          // 4.1 from IF
          if (!ioP.isFinal())
            QE2 = resolutionPtr->Q2EIF(ioP, outPQ, outJ, evolTypeGsplitF);
          // 4.2 from FF
          else
            QE2 = resolutionPtr->Q2E(ioP, outJ, outPQ, evolTypeGsplitFF);
          if (QE2 < Q2large) { iUnOrd.resize(0);
                               iUnOrd.push_back(iJ);
                               iUnOrd.push_back(iP);       
                               iUnOrd.push_back(iPQ); 
                               iUnOrd.push_back(iCPPQ);
                               Q2large = QE2; }
        }
      }
      // -- Non-Emission -- END
      
    }
    // --------- Loop over all final state partons -- END
    
    // Set
    if (Q2large < pow(10.0,100.0)) { // Otherwise we didn't find anything
      *Q2min = Q2large;
      *unOrdPartons = iUnOrd;
    }
    
  }
  
  // Unordering scale for uncertainty bands (with different evolution types)
  double MEcorrections::unorderingScaleUB(vector<Particle> partonsIn, vector<double> evolTypes) {
    // Save old evolution types
    vector<double> evolTypesOld = getEvolTypes();
    // Set evolution types
    setEvolTypes(evolTypes);
    double Q2restart = 0.0;
    vector<int> iUnOrd;
    unorderingScale(partonsIn, &Q2restart, &iUnOrd);
    // Set evolution types back
    setEvolTypes(evolTypesOld);
    return Q2restart;
  }
  
  
  //************************************************************************
      
  // Find the colour chains
  
  vector< vector<int> > MEcorrections::getColChains(vector<Particle> partonsIn) {
    if (verbose >= 6) {
      printOut("Vincia::MEcorrections::colChains", "Printing momenta: id col acol p");
      for (int i=0; i<(int)partonsIn.size(); i++)
        cout << "  " << setw(2) << partonsIn[i].id() << " " << setw(3) << partonsIn[i].col()
             << " " << setw(3) << partonsIn[i].acol() << " " << partonsIn[i].p();
    }
    vector< vector<int> > cChains;
    vector<int> iQs;
    // Find and save all colours and anticolours
    map<int,int> indexOfAcol; // Put initial colours here
    map<int,int> indexOfCol;  // Put initial anticolours here
    for (int i=0; i<(int)partonsIn.size(); i++) {
      int col   = 0;
      int acol  = 0;
      bool isIS = !partonsIn[i].isFinal();
      if (isIS) {
        col  = partonsIn[i].acol();
        acol = partonsIn[i].col();
      } else {
        col  = partonsIn[i].col();
        acol = partonsIn[i].acol();
      }
      if (col > 0) indexOfCol[col] = i;
      else if (col < 0) indexOfAcol[-col] = i;
      if (acol > 0) indexOfAcol[acol] = i;
      else if (acol < 0) indexOfCol[-acol] = i;
      int id = partonsIn[i].id();
      if (abs(id) < 6) {
        if ( (id > 0) && (!isIS) ) iQs.push_back(i);
        else if ( (id < 0) && (isIS) ) iQs.push_back(i);
      }
    }
    // Gluons only, pick random one to start with
    if (iQs.size()==0) {
      int iG1 = -1;
      int col = 0;
      for (int i=0; i<(int)partonsIn.size(); i++)  if (partonsIn[iG1].id() == 21) {
        iG1 = i;
        col = (partonsIn[iG1].isFinal() ? partonsIn[iG1].col() : partonsIn[iG1].acol());
        break;
      }
      vector<int> chain; chain.clear(); chain.resize(0);
      chain.push_back(iG1);
      bool done  = false;
      int iGnext = -1;      
      while (!done) {
        iGnext = indexOfAcol[col];
        if (iGnext == iG1) done = true; // Found the end (gluon we started with)
        else {
          chain.push_back(iGnext);
          col = (partonsIn[iGnext].isFinal() ? partonsIn[iGnext].col() : partonsIn[iGnext].acol()); // Update colour to look for
        }
      }
      cChains.push_back(chain);
    }
    else { // Loop over all quarks if found
      for (int iQ=0; iQ<(int)iQs.size(); iQ++) {
        vector<int> chain; chain.clear(); chain.resize(0);
        chain.push_back(iQs[iQ]);
        int col = (partonsIn[iQs[iQ]].isFinal() ? partonsIn[iQs[iQ]].col() : partonsIn[iQs[iQ]].acol());
        bool done = false;
        int iPnext = -1;
        while (!done) {
          iPnext = indexOfAcol[col];
          chain.push_back(iPnext);
          int id = partonsIn[iPnext].id();
          if (partonsIn[iPnext].isFinal()) {
            if ( (id < 0) &&  (abs(id) < 6) ) done = true; // Found the end of the chain (antiquark)
            col = partonsIn[iPnext].col(); // Update colour to look for
          } else {
            if ( (id > 0) &&  (abs(id) < 6) ) done = true; // Found the end of the chain (antiquark)
            col = partonsIn[iPnext].acol(); // Update colour to look for
          }
        }
        cChains.push_back(chain);
      }
    }
    /*cout << "Found the following colour chains: " << endl;
    for (int i=0; i<(int)cChains.size(); i++) {
      cout << "Chain nr. " << i << " with partons ";
      for (int j=0; j<(int)cChains[i].size(); j++) {
        cout << " " << cChains[i][j];
      }
      cout << endl;
    }*/
    return cChains;
  }
  
  //************************************************************************
      
  // Checks if a clustering is possible
  
  bool MEcorrections::canClustQQB(vector<Particle> partonsIn, vector< vector<int> > cChains, 
                                  int iJ, int iPQ) {
    if (cChains.size() <= 0) return true; // Only forming chains if # quark pairs >= 2
    if ((int)cChains.size() <= nQLmin) return false;
    /*cout << "With the following colour chains: " << endl;
    for (int i=0; i<(int)cChains.size(); i++) {
      cout << "Chain nr. " << i << " with partons ";
      for (int j=0; j<(int)cChains[i].size(); j++) {
        cout << " " << cChains[i][j];
      }
      cout << endl;
    }*/
    // Flavour sum to check if Ws are involved
    int flavourSum = 0;
    int icChainQ1  = 0;
    int jcChainQ1  = 0;
    int icChainQ2  = 0;
    int jcChainQ2  = 0;
    // 1) Are the two quarks in the same colour chain?
    for (int i=0; i<(int)cChains.size(); i++) {
      bool foundQ1 = false;
      bool foundQ2 = false;
      for (int j=0; j<(int)cChains[i].size(); j++) {
        if (cChains[i][j] == iJ) {
          foundQ1   = true;
          icChainQ1 = i;
          jcChainQ1 = j;
        }
        else if (cChains[i][j] == iPQ) {
          foundQ2   = true;
          icChainQ2 = i;
          jcChainQ2 = j;
        }
        if (partonsIn[cChains[i][j]].id() != 21) {
          if (partonsIn[cChains[i][j]].isFinal()) flavourSum += partonsIn[cChains[i][j]].id();
          else flavourSum -= partonsIn[cChains[i][j]].id();
        }
      }
      if (foundQ1 && foundQ2) {
        //cout << iJ << " and " << iPQ << " can't be clustered," << endl;
        //cout << "as both quarks appear in the same colour chain." << endl;
        return false;
      }
    }
    // 2) For >= 3 colour chains: Has merged chain and remaining chain(s) each the
    //    same two but different flavours (e.g. u - g - ubar and d - dbar)? 
    if (cChains.size() >= 3) {
      // List of flavours
      map<int,int> flavours;
      for (int ifl=1; ifl<=5; ++ifl) flavours[ifl] = 0;
      // Flavours that form the colour chain, from merged one.
      // Do they have the same flavour?
      int flavEndQ1 = partonsIn[ cChains[icChainQ1][ cChains[icChainQ1].size()-1-jcChainQ1 ] ].idAbs();
      int flavEndQ2 = partonsIn[ cChains[icChainQ2][ cChains[icChainQ2].size()-1-jcChainQ2 ] ].idAbs();
      bool mergedSameCol = (flavEndQ1 == flavEndQ2);
      flavours[flavEndQ1]++;
      flavours[flavEndQ2]++;
      vector<bool> remainSameCol;
      remainSameCol.resize(0);
      // Check remaining colour chains
      for (int i=0; i<(int)cChains.size(); i++) if (i != icChainQ1 && i != icChainQ2) {
        int flavEnd1 = partonsIn[cChains[i][0]].idAbs();
        int flavEnd2 = partonsIn[cChains[i][cChains[i].size()-1]].idAbs();
        remainSameCol.push_back(flavEnd1 == flavEnd2);
        flavours[flavEnd1]++;
        flavours[flavEnd2]++;
      }
      // Do all remaining chains have the same flavour?
      bool allRemainSameCol = true;
      for (int i=0; i<(int)remainSameCol.size(); i++)
        if (!remainSameCol[i]) allRemainSameCol = false;
      // How many different flavours do we have?
      int nrDiffFlavours = 0;
      for (int ifl=1; ifl<=5; ++ifl) if (flavours[ifl] > 0) nrDiffFlavours++;
      // Merged two colour chains -> cChains.size()-1 left over
      if ((nrDiffFlavours == (int)cChains.size()-1) 
          && mergedSameCol && allRemainSameCol) {
        // We only have colour chains that each have the same quark 
        // flavour and all with different flavours -> disconnected
        //cout << iJ << " and " << iPQ << " can't be clustered," << endl;
        //cout << "as we get disconnected colour chains." << endl;
        return false;
      } else if (nrDiffFlavours == (int)cChains.size()) {
        // This can happen for W events
        //cout << iJ << " and " << iPQ << " can't be clustered," << endl;
        //cout << "as we get disconnected colour chains in W events." << endl;
        return false;
      }
    }
    //cout << iJ << " and " << iPQ << " can be clustered." << endl;
    return true;
  }

}
